import numpy as np
import arithmetic

def smooth(data,smooth,smoothtype='gaussian',downsample=True,downsample_factor=None,
        convmode='same'):
    """
    Smooth and downsample the data array

    *smooth* [ float ] 
        Number of pixels to smooth by

    *smoothtype* [ 'gaussian','hanning', or 'boxcar' ]
        type of smoothing kernel to use

    *downsample* [ bool ]
        Downsample the data?

    *downsample_factor* [ int ]
        Downsample by the smoothing factor, or something else?

    *convmode* [ 'full','valid','same' ]
        see `numpy.convolve`.  'same' returns an array of the same length as
        'data' (assuming data is larger than the kernel)
    """
    
    roundsmooth = round(smooth) # can only downsample by integers

    if downsample_factor is None and downsample:
        downsample_factor = int(roundsmooth)
    elif downsample_factor is None:
        downsample_factor = 1

    if smooth > len(data) or downsample_factor > len(data):
        raise ValueError("Error: trying to smooth by more than the spectral length.")

    if smoothtype == 'hanning':
        kernel = np.hanning(2+roundsmooth)/np.hanning(2+roundsmooth).sum()
    elif smoothtype == 'gaussian':
        xkern  = np.linspace(-5*smooth,5*smooth,smooth*11)
        kernel = np.exp(-xkern**2/(2*(smooth/np.sqrt(8*np.log(2)))**2))
        kernel /= kernel.sum()
        if len(kernel) > len(data):
            lengthdiff = len(kernel)-len(data)
            if lengthdiff % 2 == 0: # make kernel same size as data
                kernel = kernel[lengthdiff/2:-lengthdiff/2]
            else: # make kernel 1 pixel smaller than data but still symmetric
                kernel = kernel[lengthdiff/2+1:-lengthdiff/2-1]
    elif smoothtype == 'boxcar':
        kernel = np.ones(roundsmooth)/float(roundsmooth)

    # deal with NANs or masked values
    if hasattr(data,'mask'):
        if type(data.mask) is np.ndarray:
            OK = True - data.mask
            data = arithmetic._interp(np.arange(len(data)),np.arange(len(data))[OK],data[OK])
    if np.any(True - np.isfinite(data)):
        OK = np.isfinite(data)
        data = arithmetic._interp(np.arange(len(data)),np.arange(len(data))[OK],data[OK])

    if np.any(True - np.isfinite(data)):
        raise ValueError("NANs in data array after they have been forcibly removed.")

    smdata = np.convolve(data,kernel,convmode)[::downsample_factor]

    return smdata

def smooth_multispec(data,smoothfactor,**kwargs):
    """
    Smooth multiple spectra as from an ObsBlock (shape should be [speclen, nspec])
    """

    nobs = data.shape[1]

    newdata = np.array( [smooth(D,smoothfactor,**kwargs) for D in data.swapaxes(0,1)]).swapaxes(0,1)

    return newdata



import numpy as np
import itertools, cosmology

"""
To test:

import spectrum
spec = spectrum.Spectrum('sample_sdss.txt')
spec.plotter(xmin = 6400, xmax = 6800)
spec.specfit(guesses = [20, 6718.29, 5, 100, 6564.614, 20, 50, 6585.27, 20, 20, 6732.67, 5, 50, 6549.86, 5])
spec.measure()
"""

cm_per_mpc = 3.08568e+24

class Measurements(object):
    def __init__(self, Spectrum, z=None, d=None, xunits=None, fluxnorm=None,
            miscline=None, misctol=10, ignore=None, derive=True, debug=False):
        """
        This can be called after a fit is run.  It will inherit the specfit
        object and derive as much as it can from modelpars.  Just do:
        spec.measure(z, xunits, fluxnorm)
        
        Notes: If z (redshift) or d (distance) are present, we can compute
        integrated line luminosities rather than just fluxes.  Provide distance
        in cm.
        
        Currently will only work with Gaussians.  To generalize:
            1. make sure we manipulate modelpars correctly, i.e. read in
            entries corresponding to wavelength/frequency/whatever correctly.
            
        misclines = dictionary
            miscline = {{'name': H_alpha', 'wavelength': 6565, 'etc': 0}, {}}
            
        """
        self.debug = debug
                    
        # Inherit specfit object    
        self.specfit = Spectrum.specfit
        self.speclines = Spectrum.speclines
                        
        # Bit of a hack - help identifying unmatched lines
        self.miscline = miscline        
        self.misctol = misctol
                
        # Flux units in case we are interested in line luminosities or just having real flux units
        if fluxnorm is not None: self.fluxnorm = fluxnorm
        else: self.fluxnorm= 1
                
        # This is where we'll keep our results                        
        self.lines = {}
        
        # Read in observed wavelengths
        tmp1 = np.reshape(self.specfit.modelpars, (len(self.specfit.modelpars) / 3, 3))
        tmp2 = np.reshape(self.specfit.modelerrs, (len(self.specfit.modelerrs) / 3, 3))
        
        if ignore is not None:
            tmp1 = np.delete(tmp1, ignore, 0)
            tmp2 = np.delete(tmp2, ignore, 0)
                
        order = np.argsort(zip(*tmp1)[1])
        self.obspos = np.sort(list(zip(*tmp1)[1]))
        self.Nlines = len(self.obspos)
                
        # Read in modelpars and modelerrs, re-organize so they are 2D arrays sorted by ascending wavelength
        self.modelpars = np.zeros_like(tmp1)
        self.modelerrs = np.zeros_like(tmp2)
        for i, element in enumerate(order): 
            self.modelpars[i] = tmp1[element]
            self.modelerrs[i] = tmp2[element]
                                                                      
        # Read in appropriate list of reference wavelengths/frequencies/whatever
        self.reflines = self.speclines.optical.optical_lines
        self.refpos = self.reflines['xarr']
        self.refname = self.reflines['name']
                        
        # If distance or redshift has been provided, we can compute luminosities from fluxes
        if d is not None: self.d = d
        else: self.d = None
        if z is not None: 
            self.cosmology = cosmology.Cosmology()
            self.d = self.cosmology.LuminosityDistance(z) * cm_per_mpc
            self.redshift = z
        else:
            self.redshift = 0.0
            
        self.identify()
        if derive: self.derive()
        
    def identify(self):
        """
        Determine identity of lines in self.modelpars.  Fill entries of self.lines dictionary.
        
        Note: This method will be infinitely slow for more than 10 or so lines.
        """    
        
        # List to keep track of line identification.  Each entry is (cost, (line1, line2, lin3,...))
        self.IDresults = []
        
        # Spacing between observed lines (odiff) and reference lines (rdiff)
        self.odiff = np.abs(np.diff(self.obspos))
        self.rdiff = np.abs(np.diff(self.refpos))
        self.rdmin = 0.5 * min(self.rdiff)
        
        # If lines have multiple components...
        if np.any(self.odiff) < self.rdmin:
            where = np.ravel(np.argwhere(self.odiff < self.rdmin))
            odiff = np.delete(self.odiff, where)
            multi = True
        else: 
            where = 0
            odiff = self.odiff
            multi = False

        # need to account for redshift if self.redshift is set
        # WRONG! Assume REST frame, do shifting elsewhere...
        refpos = self.refpos #* (1.0+self.redshift)
        refname = self.refname     
                            
        # Don't include elements of reference array that are far away from the observed lines (speeds things up)
        condition = (refpos >= 0.9 * min(self.obspos)) & (refpos <= 1.1 * max(self.obspos)) 
        refpos = refpos[condition]
        refname = refname[condition]
                        
        combos = itertools.combinations(refpos, self.Nlines - len(where))        
        for i, combo in enumerate(combos):
            rdiff = np.diff(combo)
            self.IDresults.append((np.sum(np.abs(odiff - rdiff)), combo))
                        
        # Pick best solution
        MINloc = np.argmin(zip(*self.IDresults)[0])  # Location of best solution
        ALLloc = []                                  # x-values of best fit lines in reference dictionary
                                
        # Determine indices of matched reference lines       
        for element in self.IDresults[MINloc][1]: 
            ALLloc.append(np.argmin(np.abs(refpos - element)))                    
        
        # Fill lines dictionary            
        for i, element in enumerate(ALLloc): 
            line = refname[element]
            self.lines[line] = {}
            loc = np.argmin(np.abs(self.obspos - refpos[element]))                
            self.lines[line]['modelpars'] = list(self.modelpars[loc])            
            self.lines[line]['modelerrs'] = list(self.modelerrs[loc])            
                                                               
        # Track down odd lines (i.e. broad components of lines already identified)
        # This won't yet work for lines that are truly unidentified
        if len(ALLloc) < self.Nlines:
                                                
            # Eliminate all modelpars/errs that belong to lines that were identified
            tmp1 = list(np.ravel(self.modelpars))
            tmp2 = list(np.ravel(self.modelerrs))
            for key in self.lines.keys():
                for element in self.lines[key]['modelpars']: 
                    loc = np.argmin(np.abs(element - tmp1))
                    tmp1 = np.delete(tmp1, loc)
                    tmp2 = np.delete(tmp2, loc)
             
            # Loop over unmatched modelpars/errs, find name of unmatched line, extend corresponding dict entry
            if self.miscline is None:                                          
                try:  
                    for i, x in enumerate(zip(*tmp1)[1]):    
                        loc = np.argmin(np.abs(ALLloc - x))
                        line = refname[loc]
                        self.lines[line]['modelpars'].extend(tmp1[i:i+3])
                        self.lines[line]['modelerrs'].extend(tmp2[i:i+3])
                except TypeError:
                    loc = np.argmin(np.abs(tmp1[1] - refpos))                       
                    line = refname[loc]
                    self.lines[line]['modelpars'].extend(tmp1)
                    self.lines[line]['modelerrs'].extend(tmp2)
            
            # If we've know a-priori which lines the unmatched lines are likely to be, use that information        
            else:
                
                for i, miscline in enumerate(self.miscline):
                    try:  
                        for j, x in enumerate(zip(*tmp1)[1]):    
                            if abs(x - self.lines[miscline['name']]['modelpars'][1]) < self.misctol[i]:
                                self.lines[line]['modelpars'].extend(tmp1[j:j+3])
                                self.lines[line]['modelerrs'].extend(tmp2[j:j+3])
                    except TypeError:
                        if abs(tmp1[1] - self.lines[self.miscline[0]]['modelpars'][1]) < self.misctol:
                            self.lines[self.miscline[0]]['modelpars'].extend(tmp1)
                            self.lines[self.miscline[0]]['modelerrs'].extend(tmp2)
                            break #?
        
        
                                              
        self.separate() 
                            
    def derive(self):
        """
        Calculate luminosity and FWHM for all spectral lines.
        """            
        
        for line in self.lines.keys():
            if self.debug:
                print "Computing parameters for line %s" % line
            
            self.lines[line]['fwhm'] = self.compute_fwhm(self.lines[line]['modelpars'])
            self.lines[line]['flux'] = self.compute_flux(self.lines[line]['modelpars'])
            self.lines[line]['amp'] = self.compute_amplitude(self.lines[line]['modelpars'])
            self.lines[line]['pos'] = self.lines[line]['modelpars'][1]
            
            if self.d is not None:
                self.lines[line]['lum'] = self.compute_luminosity(self.lines[line]['modelpars'])            
                
    def separate(self):
        """
        For multicomponent lines, separate into broad and narrow components (assume only one of components is narrow).
        """
        
        for key in self.lines.keys():
            modpars = self.lines[key]['modelpars']
            moderrs = self.lines[key]['modelerrs']
            if len(modpars) > 3:
                modpars2d = np.reshape(modpars, (len(modpars) / 3, 3))
                moderrs2d = np.reshape(moderrs, (len(moderrs) / 3, 3))
                sigma = zip(*modpars2d)[2]
                minsigma = min(np.abs(sigma))
                i_narrow = list(np.abs(sigma)).index(minsigma)
            else: continue
                        
            self.lines["{0}_N".format(key)] = {}         
            self.lines["{0}_N".format(key)]['modelpars'] = []   
            self.lines["{0}_N".format(key)]['modelerrs'] = []   
            self.lines["{0}_B".format(key)] = {}            
            self.lines["{0}_B".format(key)]['modelpars'] = [] 
            self.lines["{0}_B".format(key)]['modelerrs'] = [] 
                        
            for i, arr in enumerate(modpars2d):
                if i == i_narrow: 
                    self.lines["{0}_N".format(key)]['modelpars'] = arr
                    self.lines["{0}_N".format(key)]['modelerrs'] = moderrs2d[i]
                else: 
                    self.lines["{0}_B".format(key)]['modelpars'].extend(arr)
                    self.lines["{0}_B".format(key)]['modelerrs'].extend(moderrs2d[i])
                    
    def compute_flux(self, pars):                                                                       
        """                                                                                                
        Calculate integrated flux of emission line.  Works for multi-component fits too.  Unnormalized.    
        """                                                                                                
                                                                                                           
        flux = 0                                                                                           
        for i in xrange(len(pars) / 3): flux += np.sqrt(2. * np.pi) * pars[3 * i] * abs(pars[2 + 3 * i])
                                                                                                           
        return flux * self.fluxnorm
        
    def compute_amplitude(self, pars):
        """
        Calculate amplitude of emission line.  Should be easy - add multiple components if they exist.
        Currently assumes multiple components have the same centroid.
        """
        
        amp = 0
        for i in xrange(len(pars) / 3): amp += pars[3 * i]
        return amp * self.fluxnorm
                                                                                                           
    def compute_luminosity(self, pars):                                                                 
        """                                                                                                
        Determine luminosity of line (need distance and flux units).                                                      
        """                                                                                                
                                                                                                           
        lum = 0                                                                                            
        for i in xrange(len(pars) / 3): lum += self.compute_flux(pars) * 4. * np.pi * self.d**2                   
        return lum                                                                                         
        
    def compute_fwhm(self, pars):
        """
        Determine full-width at half maximum for multi-component fit numerically, or analytically if line
        has only a single component.  Uses bisection technique for the former with absolute tolerance of 1e-4.
        """

        if len(pars) == 3: 
            return 2. * np.sqrt(2. * np.log(2.)) * abs(pars[2])
        else:
            atol = 1e-4
            pars2d = np.reshape(pars, (len(pars) / 3, 3))
            start = zip(*pars2d)[1][0]                    # start at central wavelength of first component
            
            # If the centroids are exactly the same for all components, we know the peak, and peak position
            if np.allclose(zip(*pars2d)[1], atol): 
                fmax = np.sum(zip(*pars2d)[0])            
                
            # Otherwise, we have to figure out where the multicomponent peak is
            else:    
                f = lambda x: self.specfit.fitter.slope(x)
                xfmax = self.bisection(f, start)
                fmax = self.specfit.fitter.n_modelfunc(pars)(np.array([xfmax, xfmax]))[0]
                            
            hmax = 0.5 * fmax    
                                
            # current height relative to half max - we want to minimize this function.  Could be asymmetric.
            f = lambda x: self.specfit.fitter.n_modelfunc(pars)(np.array([x])) - hmax                   
            xhmax1 = self.bisection(f, start)
            xhmax2 = self.bisection(f, start + (start - xhmax1))
                                                    
            return abs(xhmax2 - xhmax1)      
            
    def bisection(self, f, x_guess):
        """
        Find root of function using bisection method.  Absolute tolerance of 1e-4 is being used.
        """

        x1, x2 = self.bracket_root(f, x_guess)
        
        # Narrow bracketed range with bisection until tolerance is met
        while abs(x2 - x1) > 1e-4:
            midpt = np.mean([x1, x2])
            fmid = f(midpt)
        
            if np.sign(fmid) < 0: x1 = midpt
            else: x2 = midpt
            
            if fmid == 0.0: break
                        
        return x2    
        
    def bracket_root(self, f, x_guess, atol = 1e-4):
        """
        Bracket root by finding points where function goes from positive to negative.
        """
        
        f1 = f(x_guess)
        f2 = f(x_guess + 1)
        df = f2 - f1
                
        # Determine whether increasing or decreasing x_guess will lead us to zero
        if (f1 > 0 and df < 0) or (f1 < 0 and df > 0): sign = 1
        else: sign = -1
        
        # Find root bracketing points
        xpre = x_guess
        xnow = x_guess + sign 
        fpre = f1
        fnow = f(xnow)
        while (np.sign(fnow) == np.sign(fpre)):
            xpre = xnow
            xnow += sign * 0.1
            fpre = f(xpre)
            fnow = f(xnow)
                    
        x1 = min(xnow, xpre)
        x2 = max(xnow, xpre)
        
        if not np.all([np.sign(fpre), np.sign(fnow)]): 
            x1 -= 1e-4
            x2 += 1e-4
                                
        return x1, x2    
        
    def to_tex(self):
        """
        Write out fit results to tex format.
        """    
        
        pass
            
        




"""
===================
H2CO fitter wrapper
===================

Wrapper to fit formaldehyde spectra.  
"""
import pyspeckit
from matplotlib import pyplot 
import copy
import random

title_dict = {
    'oneone':'H$_2$CO 1$_{11}$-1$_{10}$', 
    'twotwo':'H$_2$CO 2$_{12}$-2$_{11}$',
    'threethree':'H$_2$CO 3$_{23}$-3$_{22}$'
    }

def plot_h2co(spdict,spectra, fignum=1, show_components=False, residfignum=None, **plotkwargs):
    """
    Plot the results from a multi-h2co fit
    """ 
    spectra.plotter.figure = pyplot.figure(fignum)
    spectra.plotter.axis = spectra.plotter.figure.gca()
    pyplot.clf()
    splist = spdict.values()

    for sp in splist:
        sp.xarr.convert_to_unit('km/s',quiet=True)
        if hasattr(spectra.specfit,'fitter'):
            sp.specfit.fitter = copy.copy(spectra.specfit.fitter)
            sp.specfit.modelpars = spectra.specfit.modelpars
            sp.specfit.npeaks = spectra.specfit.npeaks
            sp.specfit.fitter.npeaks = spectra.specfit.npeaks
            if spectra.specfit.modelpars is not None:
                sp.specfit.model = sp.specfit.fitter.n_modelfunc(pars=spectra.specfit.modelpars, **spectra.specfit.fitter.modelfunc_kwargs)(sp.xarr)

    if len(splist) == 2:
        axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(212) }
    elif len(splist) == 3:
        axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(223), 'threethree':pyplot.subplot(224), 'fourfour':pyplot.subplot(224) }
    elif len(splist) == 4:
        axdict = { 'oneone':pyplot.subplot(221), 'twotwo':pyplot.subplot(222), 'threethree':pyplot.subplot(223), 'fourfour':pyplot.subplot(224) }
    for linename,sp in spdict.iteritems():
        sp.plotter.axis=axdict[linename] # permanent
        sp.plotter(axis=axdict[linename],title=title_dict[linename], **plotkwargs)
        sp.specfit.Spectrum.plotter = sp.plotter
        #sp.specfit.selectregion(reset=True)
        if sp.specfit.modelpars is not None:
            sp.specfit.plot_fit(annotate=False, show_components=show_components)
    if spdict['oneone'].specfit.modelpars is not None:
        spdict['oneone'].specfit.annotate(labelspacing=0.05,prop={'size':'small','stretch':'extra-condensed'},frameon=False)

    if residfignum is not None:
        pyplot.figure(residfignum)
        pyplot.clf()
        if len(splist) == 2:
            axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(212) }
        elif len(splist) == 3:
            axdict = { 'oneone':pyplot.subplot(211), 'twotwo':pyplot.subplot(223), 'threethree':pyplot.subplot(224), 'fourfour':pyplot.subplot(224) }
        elif len(splist) == 4:
            axdict = { 'oneone':pyplot.subplot(221), 'twotwo':pyplot.subplot(222), 'threethree':pyplot.subplot(223), 'fourfour':pyplot.subplot(224) }
        for linename,sp in spdict.iteritems():
            sp.specfit.plotresiduals(axis=axdict[linename])

    spectra.axisdict = axdict
    spectra.plotter.axis = axdict['oneone']
    spectra.specfit.fitleg = spdict['oneone'].specfit.fitleg


def BigSpectrum_to_H2COdict(sp, vrange=None):
    """
    A rather complicated way to make the spdicts above given a spectrum...
    """

    spdict = {}
    for linename,freq in pyspeckit.spectrum.models.formaldehyde.central_freq_dict.iteritems():
        if vrange is not None:
            freq_test_low  = freq - freq * vrange[0]/pyspeckit.units.speedoflight_kms
            freq_test_high = freq - freq * vrange[1]/pyspeckit.units.speedoflight_kms
        else:
            freq_test_low = freq_test_high = freq

        if (sp.xarr.as_unit('Hz').in_range(freq_test_low) or
                sp.xarr.as_unit('Hz').in_range(freq_test_high)):
            spdict[linename] = sp.copy()
            spdict[linename].xarr.convert_to_unit('GHz')
            spdict[linename].xarr.refX = freq
            spdict[linename].xarr.refX_units = 'Hz'
            #spdict[linename].baseline = copy.copy(sp.baseline)
            #spdict[linename].baseline.Spectrum = spdict[linename]
            spdict[linename].specfit = sp.specfit.copy(parent=spdict[linename])
            spdict[linename].xarr.convert_to_unit('km/s')
            if vrange is not None:
                try:
                    spdict[linename].crop(*vrange, units='km/s')
                except IndexError:
                    # if the freq in range, but there's no data in range, remove
                    spdict.pop(linename)

    return spdict
                    
def plotter_override(sp, vrange=None, **kwargs):
    """
    Do plot_h2co with syntax similar to plotter()
    """

    spdict = BigSpectrum_to_H2COdict(sp, vrange=vrange)

    if len(spdict) not in (2,3,4):
        raise ValueError("Not enough lines; don't need to use the H2CO plot wrapper")

    plot_h2co(spdict, sp, **kwargs)



"""
:Author: Adam Ginsburg <adam.g.ginsburg@gmail.com> and Jordan Mirocha <mirochaj@gmail.com>
.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
"""
__all__ = ['spectrum','cubes','wrappers']
__version__ = '0.1.8'
import spectrum
import specwarnings
try:
    import cubes
except ImportError:
    specwarnings.warn( "pyspeckit.cubes module not imported - cubes requires pywcs" )
import wrappers
from wrappers import *
from cubes import *
from spectrum import *



"""
=======
Plotter
=======

.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
"""
import matplotlib
import matplotlib.pyplot
import matplotlib.figure
import itertools
from ..config import *
import numpy as np
from pyspeckit.specwarnings import warn
import copy
import widgets

interactive_help_message = """
Interactive key commands for plotter.  An additional help message may appear if
you have initiated the fitter.
'?' - bring up this message
'f' - initiate the /f/itter 
'b' - initiate the /b/aseliner 
'B' - initiate the /b/aseliner (reset the selection too)
"""

class Plotter(object):
    """
    Class to plot a spectrum
    """


    def __init__(self, Spectrum, autorefresh=True, title="", ylabel="",
            xlabel="", silent=True, plotscale=1.0, **kwargs):
        self.figure = None
        self.axis = None
        self.Spectrum = Spectrum

        # plot parameters
        self.offset = 0.0 # vertical offset
        self.autorefresh = autorefresh
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title  = title
        self.errorplot = None
        self.plotkwargs = kwargs
        self._xlim = [None,None]
        self._ylim = [None,None]
        self.debug = False

        self.keyclick = None
        self.silent = silent
        self.plotscale = plotscale

        self._xclick1 = None
        self._xclick2 = None

        self.automake_fitter_tool = False


    def _get_prop(xy, minmax):
        def getprop(self):
            if xy == 'x':
                if minmax == 'min':
                    return self._xlim[0]
                elif minmax == 'max':
                    return self._xlim[1]
            elif xy == 'y':
                if minmax == 'min':
                    return self._ylim[0]
                elif minmax == 'max':
                    return self._ylim[1]
        return getprop

    def _set_prop(xy, minmax):
        def setprop(self, value):
            if self.debug: print "Setting %s%s to %s" % (xy,minmax,value)
            if xy == 'x':
                if minmax == 'min':
                    self._xlim[0] = value
                elif minmax == 'max':
                    self._xlim[1] = value
            elif xy == 'y':
                if minmax == 'min':
                    self._ylim[0] = value
                elif minmax == 'max':
                    self._ylim[1] = value
        return setprop

    xmin = property(fget=_get_prop('x','min'),fset=_set_prop('x','min'))
    xmax = property(fget=_get_prop('x','max'),fset=_set_prop('x','max'))
    ymin = property(fget=_get_prop('y','min'),fset=_set_prop('y','min'))
    ymax = property(fget=_get_prop('y','max'),fset=_set_prop('y','max'))


    def _disconnect_matplotlib_keys(self):
        """
        Disconnected the matplotlib key-press callbacks
        """
        if self.figure is not None:
            self._mpl_key_callbacks = dict([(k,self.figure.canvas.callbacks.callbacks['key_press_event'].pop(k)) 
                    for k in self.figure.canvas.callbacks.callbacks['key_press_event'].keys()[0:1]])

    def _reconnect_matplotlib_keys(self):
        """
        Reconnect the previously disconnected matplotlib keys
        """
        if self.figure is not None and hasattr(self,'_mpl_key_callbacks'):
            self.figure.canvas.callbacks.callbacks['key_press_event'].update(self._mpl_key_callbacks)

    def __call__(self, figure=None, axis=None, clear=True, autorefresh=None, plotscale=1.0, **kwargs):
        """
        Plot a spectrum
        
        Keywords:
        figure - either a matplotlib figure instance or a figure number
            to pass into pyplot.figure.  
        axis - Alternative to figure, can pass an axis instance and use
            it as the plotting canvas
        clear - Clear the axis before plotting?
        """

        # figure out where to put the plot
        if isinstance(figure,matplotlib.figure.Figure):
            self.figure = figure
            self.axis = self.figure.gca()
        elif type(figure) is int:
            self.figure = matplotlib.pyplot.figure(figure)
            self.axis = self.figure.gca()
        elif self.figure is None:
            if isinstance(axis,matplotlib.axes.Axes):
                self.axis = axis
                self.figure = axis.figure
            else:
                self.figure = matplotlib.pyplot.figure()

        self._mpl_connect()

        if axis is not None:
            self._mpl_disconnect()
            self.axis = axis
            self.figure = axis.figure
            self._mpl_connect()
        elif len(self.figure.axes) > 0 and self.axis is None:
            self.axis = self.figure.axes[0] # default to first axis
        elif self.axis is None:
            self.axis = self.figure.gca()


        if clear and self.axis is not None: self.axis.clear()

        if autorefresh is not None:
            self.autorefresh = autorefresh

        self.plotscale = plotscale

        self.plotkwargs = kwargs

        self.plot(**kwargs)

    def _mpl_connect(self):
        if self.keyclick is None:
            self.keyclick = self.figure.canvas.mpl_connect('key_press_event',self.parse_keys)

    def _mpl_disconnect(self):
        self.figure.canvas.mpl_disconnect(self.keyclick)
        self.keyclick = None

    def _mpl_reconnect(self):
        self._mpl_disconnect()
        self._mpl_connect()

    def plot(self, offset=0.0, xoffset=0.0, color='k', linestyle='steps-mid',
            linewidth=0.5, errstyle=None, erralpha=0.2, silent=None,
            reset=True, refresh=True, use_window_limits=None, **kwargs):
        """
        Plot the spectrum!

        Tries to automatically find a reasonable plotting range if one is not set.  

        offset - vertical offset to add to the spectrum before plotting.  Useful if you
        want to overlay multiple spectra on a single plot

        color - default to plotting spectrum in black

        linestyle - histogram-style plotting
        linewidth - narrow lines are helpful when histo-plotting
        xmin/xmax/ymin/ymax - override defaults for plot range.  Once set, these parameters
        are sticky (i.e., replotting will use the same ranges)

        reset_[xy]limits - Reset the limits to "sensible defaults" 

        ypeakscale - Scale up the Y maximum value.  Useful to keep the
        annotations away from the data.

        errstyle - can be "fill", which draws partially transparent boxes around the data to show
            the error region, or "bars" which draws standard errorbars

        *reset* [ bool ]
            Reset the x/y axis limits?
        """

        if self.axis is None:
            raise Exception("You must call the Plotter class to initiate the canvas before plotting.")

        self.offset += offset

        self.label(**kwargs)
        for arg in ['title','xlabel','ylabel']:
            if arg in kwargs: kwargs.pop(arg)

        reset_kwargs = {}
        for arg in ['xmin','xmax','ymin','ymax','reset_xlimits','reset_ylimits','ypeakscale']:
            if arg in kwargs: reset_kwargs[arg] = kwargs.pop(arg)

        if use_window_limits == None and any(k in reset_kwargs for k in ('xmin','xmax','reset_xlimits')):
            use_window_limits = False

        if use_window_limits: self._stash_window_limits()

        self._spectrumplot = self.axis.plot(self.Spectrum.xarr+xoffset,
                self.Spectrum.data*self.plotscale+self.offset, color=color,
                linestyle=linestyle, linewidth=linewidth, **kwargs)

        if errstyle is not None:
            if errstyle == 'fill':
                order = -1 if self.Spectrum.xarr[-1] < self.Spectrum.xarr[0] else 1
                self.errorplot = [self.axis.fill_between(steppify(self.Spectrum.xarr[::order]+xoffset,isX=True),
                    steppify((self.Spectrum.data*self.plotscale+self.offset-self.Spectrum.error*self.plotscale)[::order]),
                    steppify((self.Spectrum.data*self.plotscale+self.offset+self.Spectrum.error*self.plotscale)[::order]),
                    facecolor=color, alpha=erralpha, **kwargs)]
            elif errstyle == 'bars':
                self.errorplot = self.axis.errorbar(self.Spectrum.xarr+xoffset, self.Spectrum.data*self.plotscale+self.offset,
                        yerr=self.Spectrum.error*self.plotscale, ecolor=color, fmt=None,
                        **kwargs)

        if use_window_limits: self._reset_to_stashed_limits()

        if silent is not None:
            self.silent = silent

        if reset:
            self.reset_limits(use_window_limits=use_window_limits, **reset_kwargs)

        if self.autorefresh and refresh: self.refresh()

    def _stash_window_limits(self):
        self._window_limits = self.axis.get_xlim(),self.axis.get_ylim()
        if self.debug: print "Stashed window limits: ",self._window_limits
    
    def _reset_to_stashed_limits(self):
        self.axis.set_xlim(*self._window_limits[0])
        self.axis.set_ylim(*self._window_limits[1])
        self.xmin,self.xmax = self._window_limits[0]
        self.ymin,self.ymax = self._window_limits[1]
        if self.debug: print "Recovered window limits: ",self._window_limits
    
    def reset_limits(self,xmin=None, xmax=None, ymin=None, ymax=None,
            reset_xlimits=True, reset_ylimits=True, ypeakscale=1.2,
            silent=None, use_window_limits=False, **kwargs):
        """
        Automatically or manually reset the plot limits
        """

        if use_window_limits:
            # this means DO NOT reset!
            self.set_limits_from_visible_window()
        else:
            if silent is not None:
                self.silent = silent

            if (self.Spectrum.xarr.max() < self.xmin or self.Spectrum.xarr.min() > self.xmax 
                    or reset_xlimits):
                if not self.silent: warn( "Resetting X-axis min/max because the plot is out of bounds." )
                self.xmin = None
                self.xmax = None
            if xmin is not None: self.xmin = xmin
            elif self.xmin is None: self.xmin=self.Spectrum.xarr.min()
            if xmax is not None: self.xmax = xmax
            elif self.xmax is None: self.xmax=self.Spectrum.xarr.max()

            xpixmin = np.argmin(np.abs(self.Spectrum.xarr-self.xmin))
            xpixmax = np.argmin(np.abs(self.Spectrum.xarr-self.xmax))
            if xpixmin>xpixmax: xpixmin,xpixmax = xpixmax,xpixmin
            elif xpixmin == xpixmax:
                if not self.silent: warn( "ERROR: the X axis limits specified were invalid.  Resetting." )
                self.reset_limits(reset_xlimits=True, ymin=ymin, ymax=ymax, reset_ylimits=reset_ylimits, ypeakscale=ypeakscale, **kwargs)
                return
            
            if (self.Spectrum.data.max() < self.ymin or self.Spectrum.data.min() > self.ymax
                    or reset_ylimits):
                if not self.silent and not reset_ylimits: warn( "Resetting Y-axis min/max because the plot is out of bounds." )
                self.ymin = None
                self.ymax = None

            if ymin is not None: self.ymin = ymin
            elif self.ymin is None: 
                try:
                    self.ymin=np.nanmin(self.Spectrum.data[xpixmin:xpixmax])+0.0
                except TypeError:
                    # this is assumed to be a Masked Array error
                    self.ymin = self.Spectrum.data[xpixmin:xpixmax].min() + 0.0
            if ymax is not None: self.ymax = ymax
            elif self.ymax is None:
                try:
                    self.ymax=(np.nanmax(self.Spectrum.data[xpixmin:xpixmax])-self.ymin) * ypeakscale + self.ymin
                except TypeError:
                    self.ymax=((self.Spectrum.data[xpixmin:xpixmax]).max()-self.ymin) * ypeakscale + self.ymin

            self.ymin += self.offset
            self.ymax += self.offset

        self.axis.set_xlim(self.xmin,self.xmax)
        self.axis.set_ylim(self.ymin,self.ymax)
        

    def label(self, title=None, xlabel=None, ylabel=None, verbose_label=False, **kwargs):
   
        if title is not None:
            self.title = title
        elif hasattr(self.Spectrum,'specname'):
            self.title = self.Spectrum.specname
        if self.title is not "":
            self.axis.set_title(self.title)

        if xlabel is not None:
            self.xlabel = xlabel
        elif isinstance(self.Spectrum.xarr.xtype,str):
            self.xlabel = self.Spectrum.xarr.xtype.title()
            if verbose_label:
                self.xlabel = "%s %s %s" % ( self.Spectrum.xarr.velocity_convention.title(),
                        self.Spectrum.xarr.frame.title(),
                        self.xlabel )
            if isinstance(self.Spectrum.xarr.units,str):
                self.xlabel += " ("+self.Spectrum.xarr.units+")"
        if self.xlabel is not None:
            self.axis.set_xlabel(self.xlabel)

        if ylabel is not None:
            self.ylabel=ylabel
            self.axis.set_ylabel(self.ylabel)
        elif self.Spectrum.units in ['Ta*','Tastar','K']:
            self.axis.set_ylabel("$T_A^*$ (K)")
        elif self.Spectrum.units == 'mJy':
            self.axis.set_ylabel("$S_\\nu$ (mJy)")
        elif self.Spectrum.units == 'Jy':
            self.axis.set_ylabel("$S_\\nu$ (Jy)")            
        else:
            if "$" in self.Spectrum.units:
                # assume LaTeX already
                self.axis.set_ylabel(self.Spectrum.units)
            elif len(self.Spectrum.units.split()) > 1: 
                if self.Spectrum.units.split()[1] in ['erg/cm^2/s/Ang',
                        'erg/cm2/s/A', 'erg/cm2/s/Ang', 'erg/cm/s/Ang']:
                    norm = parse_norm(self.Spectrum.units.split()[0])
                    self.axis.set_ylabel("$%s \\mathrm{erg/s/cm^2/\\AA}$" % norm)
                elif self.Spectrum.units.split()[1] in ['W/m^2/Hz','w/m^2/hz','W/m/hz','W/m/Hz']:
                    norm = parse_norm(self.Spectrum.units.split()[0])
                    self.axis.set_ylabel("$%s \\mathrm{W/m^2/Hz}$" % norm)
            else:
                label_units = parse_units(self.Spectrum.units)
                self.axis.set_ylabel(label_units)

    def refresh(self):
        if self.axis is not None:
            self.axis.figure.canvas.draw()

    def savefig(self,fname,bbox_inches='tight',**kwargs):
        """
        simple wrapper of maplotlib's savefig.  
        """
        self.axis.figure.savefig(fname,bbox_inches=bbox_inches,**kwargs)

    def parse_keys(self,event):
        """
        Parse key commands entered from the keyboard
        """
        if hasattr(event,'key'):
            if event.key == '?':
                print interactive_help_message
            elif event.key == 'f':
                print "\n\nFitter initiated from the interactive plotter.  Matplotlib shortcut keys ('g','l','p',etc.) are disabled.  Re-enable with 'r'"
                self._disconnect_matplotlib_keys()
                self.Spectrum.specfit(interactive=True)
                if not hasattr(self,'FitterTool') and self.automake_fitter_tool:
                    self.FitterTool = widgets.FitterTools(self.Spectrum.specfit, self.figure)
                elif hasattr(self,'FitterTool') and self.FitterTool.toolfig.number not in matplotlib.pyplot.get_fignums():
                    self.FitterTool = widgets.FitterTools(self.Spectrum.specfit, self.figure)
            elif event.key == 'b':
                print "\n\nBaseline initiated from the interactive plotter.  Matplotlib shortcut keys ('g','l','p',etc.) are disabled.  Re-enable with 'r'"
                self._disconnect_matplotlib_keys()
                self.Spectrum.baseline(interactive=True, reset_selection=False)
                if not hasattr(self,'FitterTool') and self.automake_fitter_tool:
                    self.FitterTool = widgets.FitterTools(self.Spectrum.specfit, self.figure)
                elif hasattr(self,'FitterTool') and self.FitterTool.toolfig.number not in matplotlib.pyplot.get_fignums():
                    self.FitterTool = widgets.FitterTools(self.Spectrum.specfit, self.figure)
            elif event.key == 'B':
                print "\n\nBaseline initiated from the interactive plotter (with reset).  Matplotlib shortcut keys ('g','l','p',etc.) are disabled.  Re-enable with 'r'"
                self._disconnect_matplotlib_keys()
                self.Spectrum.baseline(interactive=True, reset_selection=True)
                if not hasattr(self,'FitterTool') and self.automake_fitter_tool:
                    self.FitterTool = widgets.FitterTools(self.Spectrum.specfit, self.figure)
                elif hasattr(self,'FitterTool') and self.FitterTool.toolfig.number not in matplotlib.pyplot.get_fignums():
                    self.FitterTool = widgets.FitterTools(self.Spectrum.specfit, self.figure)
            elif event.key == 'r':
                print "\n\nReconnected matplotlib shortcut keys."
                self._reconnect_matplotlib_keys()

    def get_two_clicks(self,event):

        if self._xclick1 is None:
            self._xclick1 = event.xdata
        elif self._xclick2 is None:
            self._xclick2 = event.xdata

    def set_limits_from_visible_window(self, debug=False):
        """ Hopefully self-descriptive: set the x and y limits from the
        currently visible window (use this if you use the pan/zoom tools or
        manually change the limits) """
        if debug:
            print "Changing x limits from %f,%f to %f,%f" % (self.xmin,self.xmax,self.axis.get_xlim()[0],self.axis.get_xlim()[1])
            print "Changing y limits from %f,%f to %f,%f" % (self.ymin,self.ymax,self.axis.get_ylim()[0],self.axis.get_ylim()[1])
        self.xmin, self.xmax = self.axis.get_xlim()
        self.ymin, self.ymax = self.axis.get_ylim()
        if debug:
            print "New x limits %f,%f == %f,%f" % (self.xmin,self.xmax,self.axis.get_xlim()[0],self.axis.get_xlim()[1])
            print "New y limits %f,%f == %f,%f" % (self.ymin,self.ymax,self.axis.get_ylim()[0],self.axis.get_ylim()[1])

    def copy(self, parent=None):
        """
        Create a copy of the plotter with blank (uninitialized) axis & figure

        [ parent ] 
            A spectroscopic axis instance that is the parent of the specfit
            instance.  This needs to be specified at some point, but defaults
            to None to prevent overwriting a previous plot.
        """

        newplotter = copy.copy(self)
        newplotter.Spectrum = parent
        newplotter.axis = None
        newplotter.figure = None

        return newplotter

    def line_ids(self, line_names, line_xvals, xval_units=None, auto_yloc=True,
            auto_yloc_fraction=0.9,  **kwargs):
        """
        Add line ID labels to a plot using lineid_plot
        http://oneau.wordpress.com/2011/10/01/line-id-plot/
        https://github.com/phn/lineid_plot
        http://packages.python.org/lineid_plot/

        Parameters
        ----------
        line_names : list
            A list of strings to label the specified x-axis values
        line_xvals : list
            List of x-axis values (e.g., wavelengths) at which to label the lines
        xval_units : string
            A valid unit to convert to.  If None, leaves units unchanged
        auto_yloc : bool
            If set, overrides box_loc and arrow_tip (the vertical position of
            the lineid labels) in kwargs to be `auto_yloc_fraction` of the plot
            range
        auto_yloc_fraction: float in range [0,1]
            The fraction of the plot (vertically) at which to place labels

        Examples
        --------
        >>> import numpy as np
        >>> import pyspeckit
        >>> sp = pyspeckit.Spectrum(
                xarr=pyspeckit.units.SpectroscopicAxis(np.linspace(-50,50,101),
                    units='km/s', refX=6562.8, refX_units='angstroms'),
                data=np.random.randn(101), error=np.ones(101))
        >>> sp.plotter()
        >>> sp.plotter.line_ids(['H$\\alpha$'],[6562.8],xval_units='angstroms')
        """
        import lineid_plot

        # convert line_xvals to current units
        xvals = [self.Spectrum.xarr.x_to_coord(c, xval_units) for c in line_xvals]

        if auto_yloc:
            yr = self.axis.get_ylim()
            kwargs['box_loc'] = (yr[1]-yr[0])*auto_yloc_fraction + yr[0]
            kwargs['arrow_tip'] = (yr[1]-yr[0])*(auto_yloc_fraction*0.9) + yr[0]


        lineid_plot.plot_line_ids(self.Spectrum.xarr,
                self.Spectrum.data, 
                xvals,
                line_names,
                ax=self.axis,
                **kwargs)

    def line_ids_from_measurements(self, auto_yloc=True,
            auto_yloc_fraction=0.9, **kwargs):
        """
        Add line ID labels to a plot using lineid_plot
        http://oneau.wordpress.com/2011/10/01/line-id-plot/
        https://github.com/phn/lineid_plot
        http://packages.python.org/lineid_plot/

        Parameters
        ----------
        auto_yloc : bool
            If set, overrides box_loc and arrow_tip (the vertical position of
            the lineid labels) in kwargs to be `auto_yloc_fraction` of the plot
            range
        auto_yloc_fraction: float in range [0,1]
            The fraction of the plot (vertically) at which to place labels

        Examples
        --------
        >>> import numpy as np
        >>> import pyspeckit
        >>> sp = pyspeckit.Spectrum(
                xarr=pyspeckit.units.SpectroscopicAxis(np.linspace(-50,50,101),
                    units='km/s', refX=6562.8, refX_units='angstroms'),
                data=np.random.randn(101), error=np.ones(101))
        >>> sp.plotter()
        >>> sp.specfit(multifit=True, fittype='gaussian', guesses=[1,0,1]) # fitting noise....
        >>> sp.measure()
        >>> sp.plotter.line_ids_from_measurements()
        """
        import lineid_plot  

        if hasattr(self.Spectrum,'measurements'):
            measurements = self.Spectrum.measurements

            if auto_yloc:
                yr = self.axis.get_ylim()
                kwargs['box_loc'] = (yr[1]-yr[0])*auto_yloc_fraction + yr[0]
                kwargs['arrow_tip'] = (yr[1]-yr[0])*(auto_yloc_fraction*0.9) + yr[0]

            lineid_plot.plot_line_ids(self.Spectrum.xarr,
                    self.Spectrum.data, 
                    [v['pos'] for v in measurements.lines.values()],
                    measurements.lines.keys(),
                    ax=self.axis,
                    **kwargs)
        else:
            warn("Cannot add line IDs from measurements unless measurements have been made!")


def parse_units(labelstring):
    import re
    labelstring = re.sub("um","$\mu$m",labelstring)
    labelstring = re.sub("-1","$^{-1}$",labelstring)
    labelstring = re.sub("-2","$^{-2}$",labelstring)
    labelstring = re.sub("-3","$^{-3}$",labelstring)
    labelstring = re.sub("ergss","ergs s",labelstring)
    return labelstring
    
def parse_norm(norm):
    """
    Expected format: norm = 10E15
    """    
        
    try: 
        base, exp = norm.split('E')
    except ValueError:
        base, exp = norm.split('e')

    if float(base) == 1.0:
        norm = '10'
    else:
        norm = base    
        
    norm += '^{%s}' % exp    
    
    return norm

def steppify(arr,isX=False):
    """
    *support function*
    Converts an array to double-length for step plotting
    """
    if isX:
        interval = abs(arr[1:]-arr[:-1]) / 2.0
        newarr = np.array(zip(arr[:-1]-interval,arr[:-1]+interval)).ravel()
        newarr = np.concatenate([newarr,2*[newarr[-1]+interval[-1]]])
    else:
        newarr = np.array(zip(arr,arr)).ravel()
    return newarr




import pyfits
from .. import units
import numpy.ma as ma
import numpy as np
from . import make_axis
import operator
from pyspeckit.specwarnings import warn

def open_1d_fits(filename,**kwargs):
    """
    Grabs all the relevant pieces of a simple FITS-compliant 1d spectrum

    Inputs:
        wcstype - the suffix on the WCS type to get to
            velocity/frequency/whatever
        specnum - Which # spectrum, along the y-axis, is 
            the data?
        errspecnum - Which # spectrum, along the y-axis,
            is the error spectrum?

    """

    f = pyfits.open(filename)

    return open_1d_pyfits(f[0],**kwargs)


def open_1d_pyfits(pyfits_hdu,specnum=0,wcstype='',specaxis="1",errspecnum=None,
        autofix=True, scale_keyword=None, scale_action=operator.div,
        verbose=False, apnum=0, **kwargs):
    """
    This is open_1d_fits but for a pyfits_hdu so you don't necessarily have to
    open a fits file
    """

    # force things that will be treated as strings to be strings
    # this is primarily to avoid problems with variables being passed as unicode
    wcstype = str(wcstype)
    specaxis = str(specaxis)

    hdr = pyfits_hdu._header
    if autofix: 
        for card in hdr.ascard:
            try:
                if verbose: card.verify('fix')
                else: card.verify('silentfix')
            except pyfits.VerifyError:
                hdr.__delitem__(card.key)
    data = pyfits_hdu.data

    # search for the correct axis (may be 1 or 3, unlikely to be 2 or others)
    # 1 = 1D spectrum
    # 3 = "3D" spectrum with a single x,y point (e.g., JCMT smurf/makecube)
    if hdr.get('NAXIS') > 1:
        for ii in xrange(1,hdr.get('NAXIS')+1):
            ctype = hdr.get('CTYPE%i'%ii)
            if ctype in units.xtype_dict:
                specaxis="%i" % ii

    if hdr.get('NAXIS') == 2:
        if hdr.get('WAT0_001') is not None:
            if 'multispec' in hdr.get('WAT0_001'):
                # treat as an Echelle spectrum from  IRAF
                warn("""
This looks like an Echelle spectrum.   You may want to load it
using pyspeckit.wrappers.load_IRAF_multispec.  The file will still
be successfully read if you continue, but the plotting and fitting packages
will run into errors.""")
                return read_echelle(pyfits_hdu)

        if isinstance(specnum,list):
            # allow averaging of multiple spectra (this should be modified
            # - each spectrum should be a Spectrum instance)
            spec = ma.array(data[specnum,:]).mean(axis=0)
        elif isinstance(specnum,int):
            spec = ma.array(data[specnum,:]).squeeze()
        else:
            raise TypeError(
                "Specnum is of wrong type (not a list of integers or an integer)." +
                "  Type: %s" %
                str(type(specnum)))
        if errspecnum is not None:
            # SDSS supplies weights, not errors.    
            if hdr.get('TELESCOP') == 'SDSS 2.5-M':
                errspec = 1. / np.sqrt(ma.array(data[errspecnum,:]).squeeze())
            else:       
                errspec = ma.array(data[errspecnum,:]).squeeze()
        else:
            errspec = spec*0 # set error spectrum to zero if it's not in the data

    elif hdr.get('NAXIS') > 2:
        if hdr.get('BANDID2'):
            # this is an IRAF .ms.fits file with a 'background' in the 3rd dimension
            spec = ma.array(data[specnum,apnum,:]).squeeze()
        else:
            for ii in xrange(3,hdr.get('NAXIS')+1):
                # only fail if extra axes have more than one row
                if hdr.get('NAXIS%i' % ii) > 1:
                    raise ValueError("Too many axes for open_1d_fits")
            spec = ma.array(data).squeeze()
        if errspecnum is None: 
            errspec = spec*0 # set error spectrum to zero if it's not in the data
    else:
        spec = ma.array(data).squeeze()
        if errspecnum is None: errspec = spec*0 # set error spectrum to zero if it's not in the data

    try:
        print "Found SCALE keyword %s.  Using %s to scale it" % (scale_keyword,scale_action)
        scaleval = hdr[scale_keyword]
        spec = scale_action(spec,scaleval)
        errspec = scale_action(errspec,scaleval)
    except (ValueError, KeyError) as e:
        pass

    xarr = None
    if hdr.get('ORIGIN') == 'CLASS-Grenoble':
        # Use the CLASS FITS definition (which is non-standard)
        # http://iram.fr/IRAMFR/GILDAS/doc/html/class-html/node84.html
        # F(n) = RESTFREQ + CRVALi + ( n - CRPIXi ) * CDELTi
        if verbose: print "Loading a CLASS .fits spectrum"
        dv = -1*hdr.get('CDELT1')
        if hdr.get('RESTFREQ'):
            v0 = hdr.get('RESTFREQ') + hdr.get('CRVAL1')
        elif hdr.get('RESTF'):
            v0 = hdr.get('RESTF') + hdr.get('CRVAL1')
        else:
            warn("CLASS file does not have RESTF or RESTFREQ")
        p3 = hdr.get('CRPIX1')
    elif hdr.get(str('CD%s_%s%s' % (specaxis,specaxis,wcstype))):
        dv = hdr['CD%s_%s%s' % (specaxis,specaxis,wcstype)]
        v0 = hdr['CRVAL%s%s' % (specaxis,wcstype)]
        p3 = hdr['CRPIX%s%s' % (specaxis,wcstype)]
        try: # astropy.io.fits is not backwards compatible
            hdr.update('CDELT%s' % specaxis,dv)
        except AttributeError:
            hdr.set('CDELT%s' % specaxis,dv)
        if verbose: print "Using the FITS CD matrix.  PIX=%f VAL=%f DELT=%f" % (p3,v0,dv)
    elif hdr.get(str('CDELT%s%s' % (specaxis,wcstype))):
        dv = hdr['CDELT%s%s' % (specaxis,wcstype)]
        v0 = hdr['CRVAL%s%s' % (specaxis,wcstype)]
        p3 = hdr['CRPIX%s%s' % (specaxis,wcstype)]
        if verbose: print "Using the FITS CDELT value.  PIX=%f VAL=%f DELT=%f" % (p3,v0,dv)
    elif len(data.shape) > 1:
        if verbose: print "No CDELT or CD in header.  Assuming 2D input with 1st line representing the spectral axis."
        # try assuming first axis is X axis
        if hdr.get('CUNIT%s%s' % (specaxis,wcstype)):
            xarr = data[0,:]
            spec = data[1,:]
            if data.shape[0] > 2:
                errspec = data[2,:]
        else:
            raise TypeError("Don't know what type of FITS file you've input; "+
                "its header is not FITS compliant and it doesn't look like it "+
                "was written by pyspeckit.")

    # Deal with logarithmic wavelength binning if necessary
    if xarr is None:
        if hdr.get('WFITTYPE') == 'LOG-LINEAR':
            xconv = lambda v: 10**((v-p3+1)*dv+v0)
            xarr = xconv(np.arange(len(spec)))
        else:
            xconv = lambda v: ((v-p3+1)*dv+v0)
            xarr = xconv(np.arange(len(spec)))
    
    restfreq = hdr.get('RESTFREQ')
    if restfreq is None: restfrq= hdr.get('RESTFRQ')

    XAxis = make_axis(xarr,hdr,wcstype=wcstype,specaxis=specaxis,**kwargs)

    return spec,errspec,XAxis,hdr

def read_echelle(pyfits_hdu):
    """
    Read an IRAF Echelle spectrum
    
    http://iraf.noao.edu/iraf/ftp/iraf/docs/specwcs.ps.Z
    """

    hdr = pyfits_hdu.header

    WAT1_dict = dict( [s.split('=') for s in hdr.get("WAT1_001").split()] )
    # hdr.get does not preserve whitespace, but whitespace is ESSENTIAL here!
    WAT_string = ""
    ii = 0
    while (hdr.get("WAT2_%03i" % (ii+1))) is not None:
        WAT_string += (hdr.get("WAT2_%03i" % (ii+1))
                + " "*(68-len(hdr.get("WAT2_%03i" % (ii+1)))))
        ii += 1

    WAT_list = WAT_string.split("spec")

    if "multi" in WAT_list[0]:
        WAT_list.pop(0)
    if ' ' in WAT_list:
        WAT_list.remove(' ')
    if '' in WAT_list:
        WAT_list.remove('')

    x_axes = []

    specaxdict = dict( [ (int(s.split("=")[0]),s.split("=")[1]) for s in WAT_list ] )
    for specnum, axstring in specaxdict.iteritems():
        axsplit = axstring.replace('"','').split()
        if specnum != int(axsplit[0]):
            raise ValueError("Mismatch in IRAF Echelle specification")
        num,beam,dtype,crval,cdelt,naxis,z,aplow,aphigh = axsplit[:9]
        if len(axsplit) > 9:
            functions = axsplit[9:]

        if int(dtype) == 0:
            xax = ( float(crval) + float(cdelt) * (np.arange(int(naxis)) + 1) ) / (1.+float(z))

        headerkws = {'CRPIX1':1, 'CRVAL1':crval, 'CDELT1':cdelt,
                'NAXIS1':naxis, 'NAXIS':1, 'REDSHIFT':z,
                'CTYPE1':'wavelength', 'CUNIT1':WAT1_dict['units']}
        cards = [pyfits.Card(k,v) for (k,v) in headerkws.iteritems()]
        header = pyfits.Header(cards)

        xarr = make_axis(xax,header)
        x_axes.append(xarr)
    
    return pyfits_hdu.data, pyfits_hdu.data*0, units.EchelleAxes(x_axes), hdr



"""
====================
Voigt Profile Fitter
====================
"""
import model
import numpy
try:
    import scipy.special
    scipyOK = True
except ImportError:
    scipyOK = False

def voigt(xarr,amp,xcen,Gfwhm,Lfwhm):
    """
    voigt profile

    V(x,sig,gam) = Re(w(z))/(sig*sqrt(2*pi))
    z = (x+i*gam)/(sig*sqrt(2))

    Converted from 
    http://mail.scipy.org/pipermail/scipy-user/2011-January/028327.html
    """

    if scipyOK:
        tmp = 1.0/scipy.special.wofz(numpy.zeros((len(xarr))) \
              +1j*numpy.sqrt(numpy.log(2.0))*Lfwhm).real
        tmp = tmp*amp* \
              scipy.special.wofz(2*numpy.sqrt(numpy.log(2.0))*(xarr-xcen)/Gfwhm+1j* \
              numpy.sqrt(numpy.log(2.0))*Lfwhm).real
        return tmp
    else:
        raise ImportError("Couldn't import scipy, therefore cannot do voigt profile stuff")

def voigt_fitter(multisingle='multi'):
    """
    Generator for voigt fitter class
    """

    myclass =  model.SpectralModel(voigt, 4,
            parnames=['amplitude','shift','gwidth','lwidth'], 
            parlimited=[(False,False),(False,False),(True,False),(True,False)], 
            parlimits=[(0,0), (0,0), (0,0), (0,0)],
            shortvarnames=('A',r'\Delta x',r'\sigma_G',r'\sigma_L'),
            multisingle=multisingle,
            )
    myclass.__name__ = "voigt"
    
    return myclass



import atpy
import urllib,urllib2
import tempfile

"""
Module to search Splatalogue.net via splat, modeled loosely on
ftp://ftp.cv.nrao.edu/NRAO-staff/bkent/slap/idl/
"""

length_dict = {'meters':1.0,'m':1.0,
        'centimeters':1e-2,'cm':1e-2,
        'millimeters':1e-3,'mm':1e-3,
        'nanometers':1e-9,'nm':1e-9,
        'micrometers':1e-6,'micron':1e-6,'microns':1e-6,'um':1e-6,
        'kilometers':1e3,'km':1e3,
        'angstroms':1e-10,'A':1e-10,
        }

# example query of SPLATALOGUE directly:
# http://www.cv.nrao.edu/php/splat/c.php?sid%5B%5D=64&sid%5B%5D=108&calcIn=&data_version=v2.0&from=&to=&frequency_units=MHz&energy_range_from=&energy_range_to=&lill=on&tran=&submit=Search&no_atmospheric=no_atmospheric&no_potential=no_potential&no_probable=no_probable&include_only_nrao=include_only_nrao&displayLovas=displayLovas&displaySLAIM=displaySLAIM&displayJPL=displayJPL&displayCDMS=displayCDMS&displayToyaMA=displayToyaMA&displayOSU=displayOSU&displayRecomb=displayRecomb&displayLisa=displayLisa&displayRFI=displayRFI&ls1=ls1&ls5=ls5&el1=el1

def query_splatalogue(minwav=0.00260,maxwav=0.00261,
        waveunits='m',root_url='http://find.nrao.edu/splata-slap/slap',
        chemical_element=None, NRAO_Recommended=True,):
    """
    Acquire an atpy table of a splatalogue searched based on wavelength.

    Future work will allow queries based on other parameters.  I'm waiting on
    development by the SPLATASLAP database folks to implement these.
    """

    if waveunits in length_dict:
        minwav = minwav * length_dict[waveunits]
        maxwav = maxwav * length_dict[waveunits]

    #query_url = "%s?REQUEST=queryData&WAVELENGTH=%f/%f" % (root_url,minwav,maxwav)
    # This is probably the more robust/pythonic way to do this sort of query:
    request_dict = {"REQUEST":"queryData","WAVELENGTH":"%f/%f" % (minwav,maxwav)}
    if chemical_element is not None: request_dict['CHEMICAL_ELEMENT'] = chemical_element
    #if NRAO_Recommended: request_dict['include_only_nrao'] = "yes"
    if NRAO_Recommended: request_dict['recommended'] = "1"
    #query_url = urllib2.Request(url=root_url,
    #        data=urllib.urlencode(request_dict))
    query_url = "%s?%s" % (root_url,urllib.urlencode(request_dict))

    U = urllib2.urlopen(query_url)

    # Normally it would be possible to do this:
    # t = atpy.Table(U,type='vo')
    # instead we have to write to a file and flush it. 
    # (see the error message below)

    R = U.read()
    U.close()
    tf = tempfile.NamedTemporaryFile()
    #for line in R:
    #    print >>tf,line.strip()
    print >>tf,R
    print tf.name
    tf.file.flush()
    t = atpy.Table(tf.name,type='vo')

    return t

"""
U = urllib2.urlopen(query_url)
t = atpy.Table(U,type='vo')
None:9:0: W35: 'value' attribute required for 'INFO' elements
None:18:0: W03: Implicitly generating an ID from a name 'catalog name' -> 'catalog_name'
None:27:0: W03: Implicitly generating an ID from a name 'molecular formula' -> 'molecular_formula'
None:33:0: W03: Implicitly generating an ID from a name 'molecule type' -> 'molecule_type'
None:54:0: W03: Implicitly generating an ID from a name 'frequency recommended' -> 'frequency_recommended'
None:57:0: W03: Implicitly generating an ID from a name 'quantum numbers' -> 'quantum_numbers'
------------------------------------------------------------
Traceback (most recent call last):
  File "<ipython console>", line 1, in <module>
  File "/Library/Frameworks/Python.framework/Versions/2.6/lib/python2.6/site-packages/atpy/basetable.py", line 167, in __init__
    self.read(*args, **kwargs)
  File "/Library/Frameworks/Python.framework/Versions/2.6/lib/python2.6/site-packages/atpy/basetable.py", line 213, in read
    atpy._readers[table_type](self, *args, **kwargs)
  File "/Library/Frameworks/Python.framework/Versions/2.6/lib/python2.6/site-packages/atpy/votable.py", line 86, in read
    votable = parse(filename, pedantic=pedantic)
  File "/Library/Frameworks/Python.framework/Versions/2.6/lib/python2.6/site-packages/vo/table.py", line 103, in parse
    return tree.VOTableFile(config=config, pos=(1, 1)).parse(iterator, config)
  File "/Library/Frameworks/Python.framework/Versions/2.6/lib/python2.6/site-packages/vo/tree.py", line 2921, in parse
    for start, tag, data, pos in iterator:
ValueError: 1:0: no element found
"""




class ParinfoList(list):
    """
    Store a list of model parameter values and their associated metadata (name,
    error, order, limits, etc.) in a class-friendly manner
    """
    def __init__(self, *args):

        try:
            from lmfit import Parameters,Parameter
            list.__init__(self,[])
            if len(args) == 1 and isinstance(args[0],Parameters):
                self._from_Parameters(args[0])
                self._dict = dict([(pp['parname'],pp) for pp in self])
                return
        except ImportError:
            pass

        list.__init__(self, *args)

        for ii,pp in enumerate(self):
            if pp.n != ii:
                pp.n = ii

        self._check_names()
        self._set_attributes()
        self._dict = dict([(pp['parname'],pp) for pp in self])

    def __str__(self):
        return "\n".join([repr(p) for p in self])

    def _getter(attributename):
        def getattribute(self):
            return [v[attributename] for v in self]
        return getattribute

    def _setter(attributename):
        def setattribute(self, values):
            if len(values) == len(self):
                for parinf,newval in zip(self,values):
                    parinf[attributename] = newval
            else:
                raise ValueError("Must have len(new values) = %i (was %i)" % (len(self),len(values)))
        return setattribute

    def keys(self):
        """ Dictionary-like behavior """
        return self.parnames

    def items(self):
        """ Dictionary-like behavior """
        return zip(self.parnames, self[:])

    #def values(self):
    #    """ Dictionary-like behavior """
    #    return [v['value'] for v in self]


    names = property(fget=_getter('parname'), fset=_setter('parname'))
    parnames=names
    shortnames = property(fget=_getter('shortparname'), fset=_setter('shortparname'))
    shortparnames=shortnames
    values = property(fget=_getter('value'), fset=_setter('value'))
    errors = property(fget=_getter('error'), fset=_setter('error'))
    n = property(fget=_getter('n'), fset=_setter('n'))
    order=n
    fixed = property(fget=_getter('fixed'), fset=_setter('fixed'))
    limits = property(fget=_getter('limits'), fset=_setter('limits'))
    limited = property(fget=_getter('limited'), fset=_setter('limited'))
    tied = property(fget=_getter('tied'), fset=_setter('tied'))

    def __getitem__(self, key):
        if type(key) is int:
            return super(ParinfoList,self).__getitem__(key)
        else:
            return self._dict[key]

    def _set_attributes(self):
        self.__dict__.update(dict([(pp['parname'],pp) for pp in self]))

    def _check_names(self):
        """
        Make sure all names are unique.  If they're not, append #'s at the end
        (but strip #'s first)
        """
        name_counter = {}
        names_stripped = [name.strip('[0-9]') for name in self.names]
        for ii,name in enumerate(names_stripped):
            if names_stripped.count(name) > 1:
                if name in name_counter:
                    name_counter[name] += 1
                    self[ii]['parname'] = self[ii]['parname'].strip('[0-9]')+ "{0}".format(name_counter[name])
                else:
                    name_counter[name] = 0
                    self[ii]['parname'] = self[ii]['parname'].strip('[0-9]')+ "{0}".format(name_counter[name])

                    # remove un-numbered versions if numbered versions are now being used
                    if name in self.__dict__:
                        self.__dict__.pop(name)

    def append(self, value):
        super(ParinfoList, self).append(value)
        self._check_names()
        self._set_attributes()

    def as_Parameters(self):
        """
        Convert a ParinfoList to an lmfit Parameters class
        """
        try:
            from lmfit import Parameters,Parameter
        except ImportError:
            print "Cannot import lmfit."
            return

        P = Parameters()
        P.add_many(*[(par.parname, par.value, not(par.fixed),
            par.limits[0] if par.limited[0] else None,
            par.limits[1] if par.limited[1] else None,
            par.tied if par.tied is not '' else None)
            for par in self])

        return P

    def _from_Parameters(self, lmpars):
        """
        Read from an lmfit Parameters instance
        """

        if len(lmpars) == len(self):
            for P in lmpars.values():
                self[P.name].value = P.value
                self[P.name].error = P.stderr
                self[P.name].limits = (P.min,P.max)
                self[P.name].limited = (P.min is not None,P.max is not None)
                self[P.name].expr = '' if P.expr is None else P.expr
        else:
            for par in lmpars.values():
                self.append(Parinfo(par))




class Parinfo(dict):
    """
    A class for storing attributes of a fitted model parameter.  It is based on
    mpfit's parinfo dictionary, which is just a dictionary containing a few set
    values.  This implements them as 'gettable' attributes instead, but far
    more importantly, includes sanity checks when setting values.

    Attributes
    ----------
    value: number
        The value of the parameter.  Arithmetic operations (*,/,+,-,**) will
        use this value
    error: number
        The error on the value
    n: int
        The order of the parameter in the model or ParinfoList
    fixed: bool
        Can the value change?  If False, error should be 0.
    limits: (min,max)
        The limits on the value of the parameter.  Only applied
        if limited
    limited: (bool, bool)
        Is the parameter value limited?
    step: number 
        from MPFIT: the step size to be used in calculating the numerical
        derivatives.  If set to zero, then the step size is computed
        automatically.  Ignored when AUTODERIVATIVE=0.
    scaleable: bool
        Is the value scaled with the data?  Important for normalization
        procedures
    tied: string
        mpfit/lmift parameter.  Allows you to specify arbitrary expressions for
        how the parameter depends on other parameters
        mpfit:
            a string expression which "ties" the parameter to other free or
            fixed parameters.  Any expression involving constants and the
            parameter array P are permitted.  Example: if parameter 2 is always
            to be twice parameter 1 then use the following: parinfo(2).tied =
            '2 * p(1)'.  Since they are totally constrained, tied parameters
            are considered to be fixed; no errors are computed for them.  
            NOTE: the PARNAME can't be used in expressions. 
    parname: string
        The parameter name
    shortparname: string (tex)
        A shortened version of the parameter name for plotting display

    """
    def __init__(self, values=None):

        dict.__init__(self, {'value':0.0, 'error':0.0,
                'n':0, 'fixed':False,
                'limits':(0.0,0.0),
                'limited':(False,False),
                'step':False,
                'scaleable':False,
                'tied':'',
                'parname':'',
                'shortparname':''})

        try:
            from lmfit import Parameter
            if isinstance(values,Parameter):
                self._from_Parameter(values)
                self.__dict__ = self
                return
        except ImportError:
            pass

        if values is not None:
            self.update(values)
        
        self.__dict__ = self

    def __repr__(self):
        try:
            reprint = "Param #%i %12s = %15g" % (self.n, self.parname, self.value)
            if self.fixed:
                reprint += " (fixed)"
            elif self.error is not None:
                reprint += " +/- %15g " % (self.error)
            if any(self.limited):
                lolim = "[%g," % self.limits[0] if self.limited[0] else "(-inf,"
                uplim = "%g]" % self.limits[1] if self.limited[1] else "inf)"
                myrange = lolim + uplim
                reprint += "  Range:%10s" % myrange
            if self.tied is not '':
                reprint += " Tied: %s" % self.tied
            if self.shortparname is not '':
                reprint += " Shortparname: %s" % self.shortparname

            return reprint
        except AttributeError:
            return super(Parinfo,self).__repr__()

    def __deepcopy__(self, memo):
        copy = Parinfo(self)
        copy.__dict__ = copy
        return copy

    def __copy__(self):
        copy = Parinfo(self)
        copy.__dict__ = copy
        return copy

    @property
    def max(self):
        return self.limits[1]

    @max.setter
    def max(self, value):
        self.limits = (self.limits[0], value)

    @property
    def min(self):
        return self.limits[0]

    @min.setter
    def min(self, value):
        self.limits = (value, self.limits[1])

    @property
    def vary(self):
        return not self.fixed

    @vary.setter
    def vary(self, value):
        self.fixed = not value

    @property
    def expr(self):
        return self.tied

    @expr.setter
    def expr(self, value):
        self._check_OK('tied',value)
        self.tied = value

    def __setattr__(self, key, value):
        # DEBUG print "Setting attribute %s = %s" % (key,value)

        self._check_OK(key,value)

        return super(Parinfo, self).__setattr__(key, value)

    def __setitem__(self, key, value):
        # DEBUG print "Setting item %s = %s" % (key,value)

        self._check_OK(key,value)

        return super(Parinfo, self).__setitem__(key, value)

    def _check_OK(self,key,value):
        # DEBUG print "Checking whether %s's value %s is OK" % (key,value)
        if key == 'value':
            if hasattr(self,'limited') and hasattr(self,'limits'):
                if self.limited[0] and value < self.limits[0]:
                    raise ValueError('Set parameter value %r < limit value %r' % (value,self.limits[0]))
                if self.limited[1] and value > self.limits[1]:
                    raise ValueError('Set parameter value %r > limit value %r' % (value,self.limits[1]))

        if key in ('limits','limited'):
            try:
                if len(value) != 2:
                    raise ValueError("%s must be a 2-tuple" % key)
            except TypeError: # if the input was scalar
                raise ValueError("%s must be a 2-tuple" % key)

        if key in ('parname','tied','shortparname'):
            if type(value) is not str:
                raise TypeError("%s must be a string" % key)

        if key in ('fixed',):
            try:
                value = bool(value)
            except:
                raise ValueError("%s had value %s, which could not be converted to boolean" % (key,value))

    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise TypeError("update expected at most 1 arguments, got %d" % len(args))
            other = dict(args[0])
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]

    def _from_Parameter(self, lmpar):
        """
        Read a Parinfo instance from and lmfit Parameter
        """
        self['limits'] = lmpar.min,lmpar.max
        self['limited'] = (lmpar.min not in (None,False),lmpar.max not in (None,False))
        self['value'] = lmpar.value
        self['error'] = lmpar.stderr
        self['parname'] = lmpar.name
        self['fixed'] = not(lmpar.vary)

    def _operation_wrapper(operation, reverse=False):
        """
        Perform an operation (addition, subtraction, mutiplication, division, etc.)
        """

        def ofunc(self, other): 
            """ Operation Function """
            intypes = type(other),type(self.value)
            try:
                returnval = getattr(self.value,'__%s__' % operation)(other)
                if type(returnval) not in intypes:
                    raise TypeError("return value had wrong type: %s" % str(type(returnval)))
                else:
                    return returnval
            except TypeError as err: # integers don't have defined operations with floats
                #print err
                #print "TypeError1: ",self, self.value, other
                try:
                    if hasattr(other,'__r%s__' % operation):
                        #print "r",operation,": ",self, self.value, other
                        return getattr(other,'__r%s__' % operation)(self.value)
                    elif hasattr(other,'__%s__' % operation[1:]):
                        #print operation,": ",self, self.value, other
                        return getattr(other,'__%s__' % operation[1:])(self.value)
                except:
                    raise TypeError("Neither side of the operation has a %s attribute!" % operation)

        return ofunc

    __add__ = _operation_wrapper('add')
    __radd__ = _operation_wrapper('radd')
    __sub__ = _operation_wrapper('sub')
    __rsub__ = _operation_wrapper('rsub')
    __mul__ = _operation_wrapper('mul')
    __rmul__ = _operation_wrapper('rmul')
    __div__ = _operation_wrapper('div')
    __rdiv__ = _operation_wrapper('rdiv')
    __pow__ = _operation_wrapper('pow')
    __rpow__ = _operation_wrapper('rpow')

    try: 
        def __array__(self):
            import numpy as np
            return np.array(self.value)
    except ImportError:
        pass


if __name__=="__main__":

    import unittest

    def check_failure(value):
        """ Raise a ValueError if value not in (0,1) """
        P = Parinfo()
        P.value = 1
        P.limited = (True,True)
        P.limits = (0,1)
        P.value = value

    def check_tied(value):
        P = Parinfo()
        P.tied = value

    def check_limits(value):
        P = Parinfo()
        P.limits = value

    def check_index(key):
        PL = ParinfoList([Parinfo({'parname':'HEIGHT'}), 
                        Parinfo({'value':15,'parname':'AMPLITUDE'}),
                        Parinfo({'value':3,'parname':'WIDTH'}),
                        Parinfo({'value':4,'parname':'WIDTH'})])
        return PL[key]

    def check_set_list(values):
        PL = ParinfoList([Parinfo({'parname':'HEIGHT'}), 
                        Parinfo({'value':15,'parname':'AMPLITUDE'}),
                        Parinfo({'value':3,'parname':'WIDTH','limits':(0,5),'limited':(True,True)}),
                        Parinfo({'value':4,'parname':'WIDTH'})])
        PL.shortparnames = ['a','b','c','d']
        PL.values = values
        return PL.values
    
    class MyTestCase(unittest.TestCase):
        def __init__(self, methodName='runTest', param=None):
            super(MyTestCase, self).__init__(methodName)
            self.param = param
        
        def test_checks_value_fail(self):
            check_failure(0.5)
            self.assertRaises(ValueError, check_failure, 5)
            self.assertRaises(ValueError, check_failure, -5)

        def test_checks_tied_fail(self):
            check_tied('p[0]')
            self.assertRaises(TypeError, check_tied, 5)
            self.assertRaises(TypeError, check_tied, (1,2,3))

        def test_checks_limits_fail(self):
            check_limits((1,2))
            self.assertRaises(ValueError, check_limits, -5)
            self.assertRaises(ValueError, check_limits, (1,2,3))

        def test_indexing(self):
            self.assertEqual(check_index(0), check_index('HEIGHT'))
            self.assertEqual(check_index(1), check_index('AMPLITUDE'))
            self.assertEqual(check_index(2), check_index('WIDTH0'))

        def test_set_list(self):
            self.assertEqual(check_set_list([1,2,3,4]),[1,2,3,4])
            self.assertRaises(ValueError,check_set_list,[1,2,10,4])

        def test_arithmetic(self):
            value = 25
            par = Parinfo({'parname':'TEST', 'value': value})
            assert par+5 == value+5
            assert par-5 == value-5
            assert par/5 == value/5
            assert par*5 == value*5

        def test_arithmetic2(self):
            value = 25.
            par = Parinfo({'parname':'TEST', 'value': value})
            assert par+5 == value+5
            assert par-5 == value-5
            assert par/5 == value/5
            assert par*5 == value*5

        def test_arithmetic3(self):
            value = 25.
            par = Parinfo({'parname':'TEST', 'value': value})
            assert par+5. == value+5.
            assert par-5. == value-5.
            assert par/5. == value/5.
            assert par*5. == value*5.

        def test_arithmetic4(self):
            value = 25
            par = Parinfo({'parname':'TEST', 'value': value})
            assert par+5. == value+5.
            assert par-5. == value-5.
            assert par/5. == value/5.
            assert par*5. == value*5.

        def test_arithmetic5(self):
            value = 25.
            par = Parinfo({'parname':'TEST', 'value': value})
            assert 5.+par == 5.+value
            assert 5.-par == 5.-value
            assert 5./par == 5./value
            assert 5.*par == 5.*value

        def test_arithmetic6(self):
            value = 25
            par = Parinfo({'parname':'TEST', 'value': value})
            assert 5.+par == 5.+value
            assert 5.-par == 5.-value
            assert 5./par == 5./value
            assert 5.*par == 5.*value

        def test_arithmetic7(self):
            value = 25.
            par = Parinfo({'parname':'TEST', 'value': value})
            assert 5+par == 5+value
            assert 5-par == 5-value
            assert 5/par == 5/value
            assert 5*par == 5*value

        def test_arithmetic8(self):
            value = 25
            par = Parinfo({'parname':'TEST', 'value': value})
            assert 5+par == 5+value
            assert 5-par == 5-value
            assert 5/par == 5/value
            assert 5*par == 5*value

        def test_copy(self):
            import copy
            value = 25
            par = Parinfo({'parname':'TEST', 'value': value})
            parcopy = copy.copy(par)
            assert parcopy.value == value
            parcopy = copy.deepcopy(par)
            assert parcopy.value == value

    unittest.main()



import pyfits
from .. import units
import numpy as np
import numpy.ma as ma

def tspec_reader(filename, merged=True, specnum=0, **kwargs):
    """ 
    Read TripleSpec reduced data

    It may be reduced by either SPEXTOOL or IRAF.

    AFAIK, Spextool will use the INSTR keyword, IRAF will use INSTRUME

    non-merged SPEXTOOL files are acceptable; they'll probably be read correctly but no guarantees

    kwargs is ignored
    """
    fitsfile = pyfits.open(filename)
    header = fitsfile[0].header
    data = fitsfile[0].data

    if merged:
        if header.get('INSTR') == 'APO Triplespec':
            # read in SPEXTOOL spectrum
            xarr = np.array(data[0,:],dtype='float64')
            spec = ma.array(data[1,:],dtype='float64')
            spec.mask = spec!=spec
            errspec = ma.array(data[2,:],dtype='float64')
            errspec.mask = (spec!=spec)+(errspec!=errspec)
            xunits = header.get('XUNITS')
            unit = header.get('YUNITS')
            header.update('BUNIT',unit)
            xtype = 'wavelength'
        elif header.get('INSTRUME') == 'tspec':
            dv,v0,p3 = header['CD1_1'],header['CRVAL1'],header['CRPIX1']
            header.update('CDELT1',dv)
            xconv = lambda v: ((v-p3+1)*dv+v0)
            xarr = xconv(np.arange(len(spec)))
            wat = dict([s.split("=") for s in header.get('WAT1_001').split()])
            xunits = wat['units']
            xtype = wat['label']
            #unit = header.get('BUNIT')
        else: # try this...
            xarr = np.array(data[specnum,0,:],dtype='float64')
            spec = ma.array(data[specnum,1,:],dtype='float64')
            spec.mask = spec!=spec
            errspec = ma.array(data[specnum,2,:],dtype='float64')
            errspec.mask = (spec!=spec)+(errspec!=errspec)
            xunits = header.get('XUNITS')
            unit = header.get('YUNITS')
            header.update('BUNIT',unit)
            xtype = 'wavelength'
    else:
        xarr = np.array(data[specnum,0,:],dtype='float64')
        spec = ma.array(data[specnum,1,:],dtype='float64')
        spec.mask = spec!=spec
        errspec = ma.array(data[specnum,2,:],dtype='float64')
        errspec.mask = (spec!=spec)+(errspec!=errspec)
        xunits = header.get('XUNITS')
        unit = header.get('YUNITS')
        header.update('BUNIT',unit)
        xtype = 'wavelength'
        
    XAxis = units.SpectroscopicAxis(xarr,xunits,xtype=xtype)

    return spec,errspec,XAxis,header



"""
showspec is my homegrown spectrum plotter, meant to somewhat follow STARLINK's
SPLAT and have functionality similar to GAIA, but with an emphasis on producing
publication-quality plots (which, while splat may do, it does unreproducibly)


TO DO:
    -add spectrum arithmetic tools
        (as is, you can use numpy.interp with sp.vind and sp.spectrum pretty
        easily)
    -implement other fitters
        -e.g., NH3 hyperfine, Voigt
    -move to object-oriented pylab/pyplot implementation (for bulk / speedup work)
    -allow for non-plotting fitting work (curious... I've never needed that yet)
    -Equivalent Width measurement without gaussian fitting
        -probably should be part of the baseline code
    -write documentation other people can read

"""



import math

import pylab
#from pylab import indices,figure,clf,savefig,plot,legend,text,axes,title,imshow,connect,get_current_fig_manager
from pylab import *
import matplotlib

from mpfit import mpfit

from collapse_gaussfit import *
from ratosexagesimal import *
import pyfits
import gaussfitter

from numpy import isnan
#from mad import MAD,nanmedian

def steppify(arr,isX=False,interval=0,sign=+1.0):
    """
    *support function*
    Converts an array to double-length for step plotting
    """
    if isX and interval==0:
        interval = abs(arr[1]-arr[0]) / 2.0
    newarr = array(zip(arr-sign*interval,arr+sign*interval)).ravel()
    return newarr

class SpecPlotter:
    """
    SpecPlotter class.  Takes in a spectrum or data cube, plotting properties,
    and a velocity axis determination function.  Look at splat_1d for a wrapper
    that might actually be useful.

    Whew, needs more documentation
    """

    def __init__(self,  cube,  axis=None,  xtol=None,  ytol=None, vconv=lambda
            x: x, xtora=lambda x: x, ytodec=lambda x: x, specname=None,
            dv=None, color='k', hdr=None, errspec=None,  maskspec=None,
            fig=None, fignum=1, clear=True, title=None, xunits='km/s',
            erralpha=0.2, ivconv=None, autorefresh=True, reffreq=None,
            gaiafignum=0, gaiafig=None, clickid=None, **kwargs ):
      self.vconv = vconv
      self.xtora = xtora
      self.ytodec = ytodec
      self.cube = cube # where(numpy.isnan(cube),0,cube)
      if len(self.cube.shape) > 1:
          self.spectrum = cube[:,0,0] # spectrum is what's plotted; cube is the "raw data"
      else:
          self.spectrum = cube # spectrum is what's plotted; cube is the "raw data"
      self.specname=specname
      self.dv=dv
      self.reffreq=reffreq
      self.scale=1.0
      self.units='K'
      self.xunits=xunits
      self.voff=0.0
      self.offset=0.0
      self.continuum=0.0
      self.errspec = errspec
      self.erralpha=erralpha
      self.plotcolor=color
      self.specfit = Specfit(self)
      self.fitspec = self.specfit
      self.baseline = Baseline(self)
      #self.fft = FFT(self)
      #self.psd = self.fft.psd
      self.vmin=None
      self.vmax=None
      self.title=title
      self.ivconv=ivconv
      self.autorefresh=autorefresh
      self.spectrumplot=None
      self.errorplot=None
      self.gaiafignum = gaiafignum
      self.gaiafig = gaiafig
      self.clickid = clickid
      self.plotkwargs = kwargs
      if maskspec is not None:
          self.maskspec = maskspec
      else:
          self.maskspec = zeros(self.cube.shape)
      self.linecollections =[]
      self.texts =[]
      if hdr: self.header = hdr
  
      # figure out where to put the plot
      if fig is None and axis is None:
          fig=figure(fignum)
          if clear: fig.clf()
          self.axis = fig.gca()
      elif fig is None and axis is None:
          self.axis = pylab.gca()
      elif fig is not None and axis is None:
          if clear: fig.clf()
          self.axis = fig.gca()
      elif fig is None and axis is not None:
          self.axis = axis
      else: # if figure and axis are both set, just use axis
          self.axis = axis
      if clear: self.axis.clear()
  
    def __call__(self, event):
      """
      Connects map cube to specplotter...
      """
      if event.inaxes:
        clickX = event.xdata
        clickY = event.ydata
        tb = get_current_fig_manager().toolbar
        #if ((self.axis is None) or (self.axis==event.inaxes)) and tb.mode=='':
        if event.button==1 and tb.mode=='':
            print "OverPlotting spectrum from point %i,%i" % (clickX,clickY)
            self.plotspec(clickY,clickX,button=event.button,cube=True)
        elif event.button==2:
            print "Plotting spectrum from point %i,%i" % (clickX,clickY)
            self.plotspec(clickY,clickX,button=event.button,cube=True,clear=True)
        elif event.button==3:
            print "Disconnecting GAIA-like tool"
            self.gaiafig.canvas.mpl_disconnect(self.clickid)
        else:
            print "Call failed for some reason: "
            print "event: ",event
  
    def plotspec(self, i=0, j=0, cube=False, title=None,
            button=1, clear=False,color=None, continuum=None,
            axis=None, offset=None, scale=None, voff=None, vmin=None,
            vmax=None, units=None, xunits=None, erralpha=None, plotpix=False,
            errstyle='fill', autorefresh=None, **kwargs):
      """
      Plot a spectrum
      Originally written to plot spectra from data cubes, hence the i,j parameter
      to specify the location in the cube
      """
  
      if kwargs.has_key('fignum'): kwargs.pop('fignum')  # HACK because I want __init__ to accept different kwargs
      if kwargs.has_key('fig'): kwargs.pop('fig')        # is there a better workaround?
      if scale  is not None: self.scale = scale
      if units  is not None: self.units = units
      if xunits is not None: self.xunits= xunits
      if voff   is not None: self.voff  = voff
      if offset is not None: self.offset= offset
      if continuum is not None: self.continuum= continuum
      if color  is not None: self.plotcolor=color
      if erralpha is not None: self.erralpha= erralpha
      if vmax  is not None: self.vmax = vmax
      if vmin  is not None: self.vmin = vmin
      if title is not None: self.title = title
      if autorefresh is not None: self.autorefresh = autorefresh
      if axis is None: axis=self.axis # allow spectrum to be plotted on other axis
      if clear: axis.clear()
  
      if plotpix:
          self.vind = arange(self.cube.shape[0])
      else:
          self.vind = self.vconv(arange(self.cube.shape[0])) + self.voff
  
      if kwargs.has_key('fignum'): kwargs.pop('fignum')
      if kwargs.has_key('linewidth'):
          linewidth = kwargs.pop('linewidth')
      else:
          linewidth=0.5
  
      if cube or len(self.cube.shape) == 3:
          self.spectrum = self.cube[:,i,j]*self.scale+self.continuum-self.baseline.basespec
          self.spectrumplot = axis.plot(self.vind,self.spectrum+self.offset,color=self.plotcolor,
                  linestyle='steps-mid',linewidth=linewidth,
                  **kwargs)
      else:
          if self.maskspec.sum() > 0:
              nanmask = where(self.maskspec,numpy.nan,1)
              self.spectrum = self.cube*self.scale*nanmask+self.continuum-self.baseline.basespec
              self.spectrumplot = axis.plot(self.vind,self.spectrum+self.offset,color=self.plotcolor,
                      linestyle='steps-mid',linewidth=linewidth,
                      **kwargs)
          else:
              self.spectrum = self.cube*self.scale+self.continuum-self.baseline.basespec
              self.spectrumplot = axis.plot(self.vind,self.spectrum+self.offset,color=self.plotcolor,
                      linestyle='steps-mid',linewidth=linewidth,
                      **kwargs)
          if self.errspec is not None:
              if errstyle == 'fill':
                  self.errorplot = [axis.fill_between(steppify(self.vind,isX=True,sign=sign(self.dv)),
                          steppify(self.spectrum+self.offset-self.errspec*self.scale),
                          steppify(self.spectrum+self.offset+self.errspec*self.scale),
                          facecolor=self.plotcolor, alpha=self.erralpha, **kwargs)]
              elif errstyle == 'bars':
                  self.errorplot = axis.errorbar(self.vind, self.spectrum+self.offset,
                          yerr=self.errspec*self.scale, ecolor=self.plotcolor, fmt=None,
                          **kwargs)
  
      if vmin is not None: xlo = self.vmin
      else: xlo=self.vind.min()
      if vmax is not None: xhi = self.vmax
      else: xhi=self.vind.max()
      axis.set_xlim(xlo,xhi)
  
      if self.title is not None:
          axis.set_title(self.title)
      elif self.xtora and self.ytodec:
          axis.set_title("Spectrum at %s %s" %
                  (ratos(self.xtora(i)),dectos(self.ytodec(j))))
      elif self.specname:
          axis.set_title("Spectrum of %s" % self.specname)
      if isinstance(self.xunits,str):
          axis.set_xlabel(self.xunits)
      else:
          axis.set_xlabel("V$_{LSR}$ (km s$^{-1}$)")
          self.xunits = 'km/s'
      if units in ['Ta*','Tastar','K']:
        axis.set_ylabel("$T_A^*$ (K)")
      elif units == 'mJy':
        axis.set_ylabel("$S_\\nu$ (mJy)")
      elif units == 'Jy':
        axis.set_ylabel("$S_\\nu$ (Jy)")
      else:
        axis.set_ylabel(self.units)
      if self.autorefresh: self.refresh()
  
    def save(self,fname,**kwargs):
      """
      Save the current spectrum (useful for saving baselined data)
      """
      newfile = pyfits.PrimaryHDU(data=self.cube,header=self.header)
      newfile.writeto(fname,**kwargs)

    def savefig(self,fname,bbox_inches='tight',**kwargs):
        """
        simple wrapper of maplotlib's savefig.  
        """
        self.axis.figure.savefig(fname,bbox_inches=bbox_inches,**kwargs)
  
    def showlines(self,linefreqs,linenames,ctype='freq',cunit='hz',yscale=0.8,vofflines=0.0,
            voffunit='km/s',**kwargs):
        """
        Overplot vertical lines and labels at the frequencies (or velocities) of each line
  
        yscale - fraction of maximum at which to label
        """

        self.clearlines() 

        if ctype != 'freq':
            print "Sorry, non-frequency units not implemented yet."
            return

        speedoflight=2.99792458e5
        if self.reffreq and self.xunits in ('km/s','m/s'):
            linefreqs = -(array(linefreqs)-self.reffreq)/self.reffreq * speedoflight
  
        if 'hz' in cunit or 'Hz' in cunit:
            linefreqs *= (1.0 + vofflines / speedoflight)
        else:
            linefreqs += vofflines
      
        ymax = (self.spectrum[self.spectrum==self.spectrum]).max() 
        for lf,ln in zip(linefreqs,linenames):
            if lf < self.vind.max() and lf > self.vind.min():
                self.linecollections.append(vlines(lf,0,ymax,**kwargs))
                self.texts.append(text(lf,ymax*yscale,ln,rotation='vertical',**kwargs))

        if self.autorefresh: self.refresh()
  
    def clearlines(self):
        if len(self.texts) > 0:
            for T in self.texts:
                if T in self.axis.texts:
                    self.axis.texts.remove(T)
        if len(self.linecollections) > 0:
            for LC in self.linecollections:
                if LC in self.axis.collections: 
                    self.axis.collections.remove(LC)
  
    def refresh(self):
        self.axis.figure.canvas.draw()

class FFT:
    def __init__(self,specplotter,fignum=3,axis=None, color='k'):
        self.specplotter=specplotter
        if axis is None:
            self.fignum=fignum
            self.figure=figure(self.fignum)
            self.axis=gca()
        else:
            self.axis=axis
            self.figure=self.axis.figure
            self.fignum=None
        #self.axis.clear()
        self.color=color
        self.fftplot=None
        self.setspec()
        self.setshift()
        self.clear()

    def __call__(self,psd=False,shift=True):
        self.setspec()
        if psd:
            self.psd(shift=shift)
        else:
            self.fft(shift=shift)
    
    def fft(self,shift=True,logplot=False,**kwargs):
        self.clear()
        self.setshift(shift)
        if logplot: self.axis.set_yscale('log')
        else: self.axis.set_yscale('linear')
        self.fftspec = fft(self.spectofft)
        self.realfft = self.fftspec.real
        self.imagfft = self.fftspec.imag
        self.fftplot = self.axis.plot(self.shiftfunc(self.realfft),
                drawstyle='steps-mid',color=self.color,**kwargs)
        self.refresh()

    def psd(self,logplot=True,shift=True,**kwargs):
        self.clear()
        if logplot: self.axis.set_yscale('log')
        else: self.axis.set_yscale('linear')
        self.setshift(shift)
        self.psdspec = fft(self.spectofft) * fft(self.spectofft[::-1])
        self.psdreal = abs(self.psdspec)
        self.fftplot = self.axis.plot(self.shiftfunc(self.psdreal),
                drawstyle='steps-mid',color=self.color,**kwargs)
        if logplot: self.axis.set_yscale('log')
        else: self.axis.set_yscale('linear')
        self.refresh()

    def setshift(self,shift=True):
        if shift: self.shiftfunc = fftshift
        else: self.shiftfunc = lambda x: x

    def setspec(self):
        self.spectofft = copy(self.specplotter.spectrum)
        OKmask = (self.spectofft==self.spectofft)
        self.spectofft[(True-OKmask)] = 0

    def clear(self):
        if self.fftplot is not None:
            for p in self.fftplot:
                p.set_visible(False)
                if p in self.axis.lines: self.axis.lines.remove(p)
        self.axis.clear()
        self.refresh()

    def refresh(self):
        self.axis.figure.canvas.draw()

class PSD(FFT):
    def __call__(self,shift=True):
        self.setspec()
        self.setshift(shift)
        self.clear()
        self.psd()
        self.refresh()

class Baseline:
    def __init__(self,specplotter):
        self.baselinepars  = None
        self.order = None
        self.basespec = zeros(specplotter.spectrum.shape[0])
        self.excludemask = zeros(specplotter.spectrum.shape[0],dtype='bool')
        self.OKmask = ones(specplotter.spectrum.shape[0],dtype='bool')
        self.specplotter = specplotter
        self.blleg = None
        self.click = 0
        self.nclicks_b1 = 0
        self.nclicks_b2 = 0
        self.fitregion=[]
        self.excludevelo = []
        self.excludepix  = []

    def __call__(self, order=1, annotate=False, excludefit=False, save=True,
            exclude=None, exclusionlevel=0.01,
            interactive=False, **kwargs):
        """
        Fit and remove a polynomial from the spectrum.  
        It will be saved in the variable "self.basespec"
        and the fit parameters will be saved in "self.order"

        function baseline(spectrum,xarr=None,xmin=None,xmax=None,order=1,quiet=True,exclude=None):
        Subtract a baseline from a spectrum
        If xmin,xmax are not specified, defaults to ignoring first and last 10% of spectrum

        exclude is a set of start/end indices to ignore when baseline fitting
        (ignored by setting error to infinite in fitting procedure)

        excludefit creates a mask based on the fitted gaussian model (assuming
        that it has a zero-height) using an exclusion level of (exclusionlevel)
        * the smallest gaussian peak that was fit

        "basespec" is added back to the spectrum before fitting so you can run this
        procedure multiple times without losing information
        """
        specfit = self.specplotter.specfit
        self.order = order
        fitp = zeros(self.order+1)
        self.spectofit = self.specplotter.spectrum+self.basespec
        self.OKmask = (self.spectofit==self.spectofit)
        if exclude == 'interactive' or interactive:
            self.excludemask[:] = True
            self.excludevelo = []
            self.excludepix  = []
            self.click = self.specplotter.axis.figure.canvas.mpl_connect('button_press_event',self.selectregion)
        else:
            if excludefit and specfit.modelpars is not None:
                #vlo = self.specplotter.specfit.modelpars[1] - 2*self.specplotter.specfit.modelpars[2]
                #vhi = self.specplotter.specfit.modelpars[1] + 2*self.specplotter.specfit.modelpars[2]
                #exclude = [argmin(abs(self.specplotter.vind-vlo)),argmin(abs(self.specplotter.vind-vhi))]
                specfit.fullsizemodel() # make sure the spectrum is the right size
                self.excludemask = abs(specfit.model) > exclusionlevel*abs(min(specfit.modelpars[0::3]))
            else:
                self.excludemask[:] = False
            self.dofit(exclude=exclude,annotate=annotate,**kwargs)
        if save: self.savefit()

    def dofit(self, exclude=None, excludeunits='velo', annotate=False,
            **kwargs):
        """
        Do the baseline fitting and save and plot the results.

        Can specify a region to exclude using velocity units or pixel units
        """
        if exclude is not None and excludeunits in ['velo','km/s']:
            if len(exclude) % 2 == 0:
                self.excludevelo = exclude
                self.excludepix = []
                for vl,vu in zip(exclude[::2],exclude[1::2]):
                    xl = argmin(abs(self.specplotter.vind-vl))
                    xu = argmin(abs(self.specplotter.vind-vu))
                    if xl > xu: xl,xu=xu,xl
                    self.excludemask[xl:xu] = True
                    self.excludepix += [xl,xu]
        elif excludeunits in ['pix','pixel','chan','channel']:
            if len(exclude) % 2 == 0:
                self.excludepix = []
                for xl,xu in zip(exclude[::2],exclude[1::2]):
                    if xl > xu: xl,xu=xu,xl
                    self.excludemask[xl:xu] = True
                    self.excludepix += [xl,xu]
        self.specplotter.spectrum, self.baselinepars = baseline(
                self.spectofit,
                xarr=self.specplotter.vind,
                order=self.order, exclude=None, 
                mask=(True-self.OKmask)+self.excludemask,
                **kwargs)
        self.basespec = poly1d(self.baselinepars)(self.specplotter.vind)
        if self.specplotter.spectrumplot is not None: 
            [self.specplotter.axis.lines.remove(p) for p in self.specplotter.spectrumplot]
        if self.specplotter.errorplot is not None: 
            [self.specplotter.axis.collections.remove(p) for p in self.specplotter.errorplot if isinstance(p,matplotlib.collections.PolyCollection)]
            [self.specplotter.axis.lines.remove(p) for p in self.specplotter.errorplot if isinstance(p,matplotlib.lines.Line2D)]
        self.specplotter.plotspec(**self.specplotter.plotkwargs)
        self.specplotter.axis.set_ylim(
                abs(self.specplotter.spectrum[self.OKmask].min())*1.1*sign(self.specplotter.spectrum[self.OKmask].min()),
                abs(self.specplotter.spectrum[self.OKmask].max())*1.1*sign(self.specplotter.spectrum[self.OKmask].max()))
        if annotate: self.annotate() # refreshes automatically
        elif self.specplotter.autorefresh: self.specplotter.refresh()

    def selectregion(self,event,annotate=False):
        """
        select regions for baseline fitting
        """
        if event.button == 1:
            if self.nclicks_b1 == 0:
                self.bx1 = argmin(abs(event.xdata-self.specplotter.vind))
                self.excludevelo += [self.specplotter.vind]
                self.excludepix  += [self.bx1]
                self.nclicks_b1 += 1
            elif self.nclicks_b1 == 1:
                self.bx2 = argmin(abs(event.xdata-self.specplotter.vind))
                self.nclicks_b1 -= 1
                if self.bx1 > self.bx2: self.bx1,self.bx2 = self.bx2,self.bx1
                self.fitregion += self.specplotter.axis.plot(
                        self.specplotter.vind[self.bx1:self.bx2],
                        self.specplotter.spectrum[self.bx1:self.bx2]+self.specplotter.offset,
                        drawstyle='steps-mid',
                        color='g',alpha=0.5)
                self.specplotter.refresh()
                self.excludemask[self.bx1:self.bx2] = False
                self.excludevelo += [self.specplotter.vind]
                self.excludepix  += [self.bx2]
        if event.button in [2,3]:
            disconnect(self.click)
            self.dofit(exclude=None,annotate=annotate)
            for p in self.fitregion:
                p.set_visible(False)
                self.specplotter.axis.lines.remove(p)
            self.fitregion=[] # I should be able to just remove from the list... but it breaks the loop...
            self.specplotter.refresh()

    def annotate(self,loc='upper left'):
        bltext = "bl: $y=$"+"".join(["$%+6.3gx^{%i}$" % (f,self.order-i)
            for i,f in enumerate(self.baselinepars)])
        #self.blleg = text(xloc,yloc     ,bltext,transform = self.specplotter.axis.transAxes)
        self.clearlegend()
        pl = matplotlib.collections.CircleCollection([0],edgecolors=['k'])
        self.blleg = self.specplotter.axis.legend(
                (pl,),
                (bltext,),loc=loc,markerscale=0.001,
                borderpad=0.1, handlelength=0.1, handletextpad=0.1
                )
        self.specplotter.axis.add_artist(self.blleg)
        if self.specplotter.autorefresh: self.specplotter.refresh()
  
    def clearlegend(self):
        if self.blleg is not None: 
            self.blleg.set_visible(False)
            if self.blleg in self.specplotter.axis.artists:
                self.specplotter.axis.artists.remove(self.blleg)
        if self.specplotter.autorefresh: self.specplotter.refresh()

    def savefit(self):
        if self.baselinepars is not None:
            for ii,p in enumerate(self.baselinepars):
                self.specplotter.header.update('BLCOEF%0.2i' % (ii),p,comment="Baseline power-law best-fit coefficient x^%i" % (self.order-ii-1))


class Specfit:

    def __init__(self,specplotter):
        self.model = None
        self.modelpars = None
        self.modelerrs = None
        self.modelplot = None
        self.guessplot = []
        self.fitregion = []
        self.ngauss = 0
        self.nclicks_b1 = 0
        self.nclicks_b2 = 0
        self.gx1 = 0
        self.gx2 = specplotter.spectrum.shape[0]
        self.guesses = []
        self.click = 0
        self.fitkwargs = {}
        self.auto = False
        self.autoannotate = True
        self.specplotter = specplotter
        self.gaussleg=None
        self.residuals=None
        self.setfitspec()
        #self.seterrspec()

    def __call__(self, interactive=False, usemoments=True, fitcolor='r',
            multifit=False, guesses=None, annotate=True, save=True,
            **kwargs):
        """
        Fit gaussians to a spectrum

        guesses = [height,amplitude,center,width]
        """
  
        self.fitcolor = fitcolor
        self.clear()

        self.ngauss = 0
        self.fitkwargs = kwargs
        if interactive:
            print "Left-click twice to select a fitting range, then middle-click twice to select a peak and width"
            self.nclicks_b1 = 0
            self.nclicks_b2 = 0
            self.guesses = []
            self.click = self.specplotter.axis.figure.canvas.mpl_connect('button_press_event',self.makeguess)
            self.autoannotate = annotate
        elif multifit:
            if guesses is None:
                print "You must input guesses when using multifit.  Also, baseline first!"
            else:
                self.guesses = guesses
                self.multifit()
                self.autoannotate = annotate
        else:
            #print "Non-interactive, 1D fit with automatic guessing"
            if self.specplotter.baseline.order is None:
                self.specplotter.baseline.order=0
                self.onedfit(usemoments=usemoments,annotate=annotate,**kwargs)
            else:
                self.onedfit(usemoments=usemoments,annotate=annotate,
                        vheight=False,height=0.0,**kwargs)
            if self.specplotter.autorefresh: self.specplotter.refresh()
        if save: self.savefit()

    def seterrspec(self,usestd=None,useresiduals=True):
        if self.specplotter.errspec is not None and not usestd:
            self.errspec = self.specplotter.errspec
        elif self.residuals is not None and useresiduals: 
            self.errspec = ones(self.spectofit.shape[0]) * self.residuals.std()
        else: self.errspec = ones(self.spectofit.shape[0]) * self.spectofit.std()

    def setfitspec(self):
        self.spectofit = copy(self.specplotter.spectrum)
        OKmask = (self.spectofit==self.spectofit)
        self.spectofit[(True-OKmask)] = 0
        self.seterrspec()
        self.errspec[(True-OKmask)] = 1e10

    def multifit(self):
        self.ngauss = len(self.guesses)/3
        self.setfitspec()
        if self.fitkwargs.has_key('negamp'): self.fitkwargs.pop('negamp')
        mpp,model,mpperr,chi2 = gaussfitter.multigaussfit(
                self.specplotter.vind[self.gx1:self.gx2], 
                self.spectofit[self.gx1:self.gx2], 
                err=self.errspec[self.gx1:self.gx2],
                ngauss=self.ngauss,
                params=self.guesses,
                **self.fitkwargs)
        self.chi2 = chi2
        self.dof  = self.gx2-self.gx1-self.ngauss*3
        self.model = model
        self.modelpars = mpp.tolist()
        self.modelerrs = mpperr.tolist()
        self.modelplot = self.specplotter.axis.plot(
                self.specplotter.vind[self.gx1:self.gx2],
                self.model+self.specplotter.offset, color=self.fitcolor, linewidth=0.5)
        self.residuals = self.spectofit[self.gx1:self.gx2] - self.model
        if self.autoannotate:
            self.annotate()
    
    def onedfit(self, usemoments=True, annotate=True, vheight=True, height=0, negamp=None,**kwargs):
        self.ngauss = 1
        self.auto = True
        self.setfitspec()
        if usemoments: # this can be done within gaussfit but I want to save them
            self.guesses = gaussfitter.onedmoments(
                    self.specplotter.vind[self.gx1:self.gx2],
                    self.spectofit[self.gx1:self.gx2],
                    vheight=vheight,negamp=negamp,**kwargs)
            if vheight is False: self.guesses = [height]+self.guesses
        else:
            if negamp: self.guesses = [height,-1,0,1]
            else:  self.guesses = [height,1,0,1]
        mpp,model,mpperr,chi2 = gaussfitter.onedgaussfit(
                self.specplotter.vind[self.gx1:self.gx2],
                self.spectofit[self.gx1:self.gx2],
                err=self.errspec[self.gx1:self.gx2],
                vheight=vheight,
                params=self.guesses,
                **self.fitkwargs)
        self.chi2 = chi2
        self.dof  = self.gx2-self.gx1-self.ngauss*3-vheight
        if vheight: 
            self.specplotter.baseline.baselinepars = mpp[:1] # first item in list form
            self.model = model - mpp[0]
        else: self.model = model
        self.residuals = self.spectofit[self.gx1:self.gx2] - self.model
        self.modelpars = mpp[1:].tolist()
        self.modelerrs = mpperr[1:].tolist()
        self.modelplot = self.specplotter.axis.plot(
                self.specplotter.vind[self.gx1:self.gx2],
                self.model+self.specplotter.offset, color=self.fitcolor, linewidth=0.5)
        if annotate:
            self.annotate()
            if vheight: self.specplotter.baseline.annotate()

    def fullsizemodel(self):
        """
        If the gaussian was fit to a sub-region of the spectrum,
        expand it (with zeros) to fill the spectrum.  You can 
        always recover the original by:
        origmodel = model[gx1:gx2]
        """

        if self.model.shape != self.specplotter.spectrum.shape:
            temp = zeros(self.specplotter.spectrum.shape)
            temp[self.gx1:self.gx2] = self.model
            self.model = temp
            self.residuals = self.spectofit - self.model

    def plotresiduals(self,fig=None,axis=None,clear=True,**kwargs):
        """
        Plot residuals of the fit.  Specify a figure or
        axis; defaults to figure(2).

        kwargs are passed to matplotlib plot
        """
        if axis is None:
            fig=figure(2)
            self.residualaxis = gca()
            if clear: self.residualaxis.clear()
        else:
            self.residualaxis = axis
            if clear: self.residualaxis.clear()
        self.residualplot = self.residualaxis.plot(self.specplotter.vind[self.gx1:self.gx2],
                self.residuals,drawstyle='steps-mid',
                linewidth=0.5, color='k', **kwargs)
        if self.specplotter.vmin is not None and self.specplotter.vmax is not None:
            self.residualaxis.set_xlim(self.specplotter.vmin,self.specplotter.vmax)
        self.residualaxis.figure.canvas.draw()

    def annotate(self,loc='upper right'):
        #text(xloc,yloc     ,"c=%g" % self.modelpars[1],transform = self.specplotter.axis.transAxes)
        #text(xloc,yloc-0.05,"w=%g" % self.modelpars[2],transform = self.specplotter.axis.transAxes)
        #text(xloc,yloc-0.10,"a=%g" % self.modelpars[0],transform = self.specplotter.axis.transAxes)
        self.clearlegend()
        pl = matplotlib.collections.CircleCollection([0],edgecolors=['k'])
        self.gaussleg = self.specplotter.axis.legend(
                tuple([pl]*3*self.ngauss),
                tuple(flatten(
                    [("c%i=%6.4g $\\pm$ %6.4g" % (jj,self.modelpars[1+jj*3],self.modelerrs[1+jj*3]),
                      "w%i=%6.4g $\\pm$ %6.4g" % (jj,self.modelpars[2+jj*3],self.modelerrs[2+jj*3]),
                      "a%i=%6.4g $\\pm$ %6.4g" % (jj,self.modelpars[0+jj*3],self.modelerrs[0+jj*3]))
                      for jj in range(self.ngauss)])),
                loc=loc,markerscale=0.01,
                borderpad=0.1, handlelength=0.1, handletextpad=0.1
                )
        self.gaussleg.draggable(True)
        self.specplotter.axis.add_artist(self.gaussleg)
        if self.specplotter.autorefresh: self.specplotter.refresh()

    def selectregion(self,event):
        if self.nclicks_b1 == 0:
            self.gx1 = argmin(abs(event.xdata-self.specplotter.vind))
            self.nclicks_b1 += 1
        elif self.nclicks_b1 == 1:
            self.gx2 = argmin(abs(event.xdata-self.specplotter.vind))
            self.nclicks_b1 -= 1
            if self.gx1 > self.gx2: self.gx1,self.gx2 = self.gx2,self.gx1
            if abs(self.gx1-self.gx2) > 3: # can't fit w/ fewer data than pars
                self.fitregion = self.specplotter.axis.plot(
                        self.specplotter.vind[self.gx1:self.gx2],
                        self.specplotter.spectrum[self.gx1:self.gx2]+self.specplotter.offset,
                        drawstyle='steps-mid',
                        color='c')
                if self.guesses == []:
                    self.guesses = gaussfitter.onedmoments(
                            self.specplotter.vind[self.gx1:self.gx2],
                            self.spectofit[self.gx1:self.gx2],
                            vheight=0)
                    self.ngauss = 1
                    self.auto = True
            else:
                print "Fitting region is too small (channels %i:%i).  Try again." % (self.gx1,self.gx2)

    def guesspeakwidth(self,event):
        if self.nclicks_b2 % 2 == 0:
            if self.auto:
                self.guesses[:2] = [event.ydata,event.xdata]
            else:
                self.guesses += [event.ydata,event.xdata,1]
                self.ngauss += 1
            self.nclicks_b2 += 1
            self.guessplot += [self.specplotter.axis.scatter(event.xdata,event.ydata,marker='x',c='r')]
        elif self.nclicks_b2 % 2 == 1:
            self.guesses[-1] = abs(event.xdata-self.guesses[-2])
            self.nclicks_b2 += 1
            self.guessplot += self.specplotter.axis.plot([event.xdata,
                2*self.guesses[-2]-event.xdata],[event.ydata]*2,
                color='r')
            if self.auto:
                self.auto = False
            if self.nclicks_b2 / 2 > self.ngauss:
                print "There have been %i middle-clicks but there are only %i gaussians" % (self.nclicks_b2,self.ngauss)
                self.ngauss += 1

    def clear(self,legend=True):
        if self.modelplot is not None:
            for p in self.modelplot:
                p.set_visible(False)
        if legend: self.clearlegend()

    def makeguess(self,event):
        if event.button == 1:
            self.selectregion(event)
        elif event.button == 2:
            self.guesspeakwidth(event)
        elif event.button == 3:
            disconnect(self.click)
            if self.ngauss > 0:
                print len(self.guesses)/3," Guesses: ",self.guesses," X channel range: ",self.gx1,self.gx2
                if len(self.guesses) % 3 == 0:
                    self.multifit()
                    for p in self.guessplot + self.fitregion:
                        p.set_visible(False)
                else: 
                    print "error, wrong # of pars"
        if self.specplotter.autorefresh: self.specplotter.refresh()

    def clearlegend(self):
        if self.gaussleg is not None: 
            self.gaussleg.set_visible(False)
            if self.gaussleg in self.specplotter.axis.artists:
                self.specplotter.axis.artists.remove(self.gaussleg)
        if self.specplotter.autorefresh: self.specplotter.refresh()
    
    def savefit(self):
        if self.modelpars is not None:
            for ii,p in enumerate(self.modelpars):
                if ii % 3 == 0: self.specplotter.header.update('AMP%1i' % (ii/3),p,comment="Gaussian best fit amplitude #%i" % (ii/3))
                if ii % 3 == 1: self.specplotter.header.update('CEN%1i' % (ii/3),p,comment="Gaussian best fit center #%i" % (ii/3))
                if ii % 3 == 2: self.specplotter.header.update('WID%1i' % (ii/3),p,comment="Gaussian best fit width #%i" % (ii/3))


def mapplot(plane,cube,vconv=lambda x: x,xtora=lambda x: x,ytodec=lambda x: x, gaiafignum=0, specfignum=1):
    
    gaiafig = figure(gaiafignum)
    gaiafig.clf()
    gaiaax = gaiafig.add_subplot(111)
    gaiaax.imshow(plane)

    sp = SpecPlotter(cube, vconv=vconv, xtora=xtora, ytodec=ytodec,
            gaiafignum=gaiafignum, fignum=specfignum, gaiafig=gaiafig)
    sp.clickid = gaiafig.canvas.mpl_connect('button_press_event',sp)
    #connect('button_press_event',sp)

 
def splat_3d(filename,xi=0,yi=0,vmin=None,vmax=None,button=1,dobaseline=False,exclude=None,
        smooth=None,smoothto=None,smoothtype='gaussian',order=1,savepre=None,**kwargs):
    """
    Inputs:
        vmin,vmax - range over which to baseline and plottransform = ax.transAxes
        exclude - (internal) range to exclude from baseline fit
    """
    dv,v0,p3,hdr,cube,xtora,ytodec,vconv,xunits,conversion_factor,units = open_3d(filename)

    if units is None: units="UNITS"
    if xunits is None: xunits="km/s"
    if conversion_factor == 0 or conversion_factor is None: conversion_factor=1.0
    sp = splat_1d(vpars=[dv, v0, p3], hdr=hdr, spec=cube[:, yi, xi],
            xtora=xtora, ytodec=ytodec, vconv=vconv, units=units,
            conversion_factor=conversion_factor, xunits=xunits, **kwargs)

    sp.cube = cube

    return sp

def gaia(filename,estimator='max',axis=0):
    f = pyfits.open(filename)
    hdr = f[0].header
    cube = f[0].data
    dv,v0,p3 = hdr.get('CD3_3'),hdr.get('CRVAL3'),hdr.get('CRPIX3')
    dr,r0,p1 = hdr.get('CD1_1'),hdr.get('CRVAL1'),hdr.get('CRPIX1')
    dd,d0,p2 = hdr.get('CD2_2'),hdr.get('CRVAL2'),hdr.get('CRPIX2')
    if dv is None: dv = hdr.get('CDELT3')
    if dr is None: dr = hdr.get('CDELT1')
    if dd is None: dd = hdr.get('CDELT2')
    xtora = lambda x: (x-p1+1)*dr+r0    # convert pixel coordinates to RA/Dec/Velocity
    ytodec = lambda y: (y-p2+1)*dd+d0
    vconv = lambda v: (v-p3+1)*dv+v0

    if axis > 0:
        cube = cube.swapaxes(0,axis)

    if estimator == 'max':
        p = where(isnan(cube),0,cube).max(axis=0)
    elif estimator == 'int':
        p = where(isnan(cube),0,cube).sum(axis=0) * dv
    elif estimator == 'intdivmax':
        cut = MAD(cube.ravel()) + nanmedian(cube.ravel())
        if cut < 0:
            cut = 0
        m = where(isnan(cube),0,cube).max(axis=0)
        i = where(isnan(cube),0,cube).sum(axis=0) * dv
        p = where(i<0,0,i)/where(m<=cut,numpy.inf,m)
    elif estimator[-5:] == ".fits":
        p = pyfits.open(estimator)[0].data

    mapplot(p,cube,vconv,xtora,ytodec)

def baseline_file(filename,outfilename,vmin=None,vmax=None,order=1,crop=False):
    f = pyfits.open(filename)
    hdr = f[0].header
    cube = f[0].data.squeeze()
    dv,v0,p3 = hdr.get('CD3_3'),hdr.get('CRVAL3'),hdr.get('CRPIX3')
    dr,r0,p1 = hdr.get('CD1_1'),hdr.get('CRVAL1'),hdr.get('CRPIX1')
    dd,d0,p2 = hdr.get('CD2_2'),hdr.get('CRVAL2'),hdr.get('CRPIX2')
    if dv is None: dv = hdr.get('CDELT3')
    if dr is None: dr = hdr.get('CDELT1')
    if dd is None: dd = hdr.get('CDELT2')
    vconv = lambda v: (v-p3+1)*dv+v0
    varr = vconv(arange(cube.shape[-1]))
    if vmin is None: argvmin = None
    else: argvmin = argmin(abs(varr-vmin))
    if vmax is None: argvmax = None
    else: argvmax = argmin(abs(varr-vmax))

    bspec,bfit = baseline(cube,vmin=argvmin,vmax=argvmax,order=order)


def baseline(spectrum,xarr=None,xmin='default',xmax='default',order=1,quiet=True,exclude=None,
        mask=None):
    """
    Subtract a baseline from a spectrum
    If xmin,xmax are not specified, defaults to ignoring first and last 10% of spectrum
    *unless* order > 1, in which case ignoring the ends tends to cause strange effects

    exclude is a set of start/end indices to ignore when baseline fitting
    (ignored by setting error to infinite in fitting procedure)
    """
    if xmin == 'default':
        if order <= 1: xmin = floor( spectrum.shape[-1]*0.1 )
        else:          xmin = 0
    elif xmin is None:
        xmin = 0
    if xmax == 'default':
        if order <= 1: xmax = ceil( spectrum.shape[-1]*0.9 )
        else:          xmax = spectrum.shape[-1]
    elif xmax is None:
        xmax = spectrum.shape[-1]
    
    pguess = [1]*(order+1)

    if xarr is None:
        xarr = indices(spectrum.shape).squeeze()

    subxarr = xarr[xmin:xmax]
    def mpfitfun(data,err):
        def f(p,fjac=None): return [0,numpy.ravel((poly1d(p)(subxarr)-data)/err)]
        return f

    err = ones(spectrum.shape)
    if exclude is not None:
        err[exclude[0]:exclude[1]] = 1e10
    if mask is not None:
        if mask.dtype.name != 'bool': mask = mask.astype('bool')
        err[mask] = 1e10
        spectrum[mask] = 0
    if (spectrum!=spectrum).sum() > 0:
        print "There is an error in baseline: some values are NaN"
        import pdb; pdb.set_trace()

    mp = mpfit(mpfitfun(spectrum[xmin:xmax],err[xmin:xmax]),xall=pguess,quiet=quiet)
    fitp = mp.params
    bestfit = poly1d(fitp)(xarr).squeeze()

    return (spectrum-bestfit),fitp

def open_3d(filename):
    f = pyfits.open(filename)
    hdr = f[0].header
    cube = f[0].data
    if len(cube.shape) == 4: cube=cube[0,:,:,:]
    #cube = reshape(cube.mean(axis=2).mean(axis=1),[cube.shape[0],1,1])
    dv,v0,p3 = hdr.get('CD3_3'),hdr.get('CRVAL3'),hdr.get('CRPIX3')
    dr,r0,p1 = hdr.get('CD1_1'),hdr.get('CRVAL1'),hdr.get('CRPIX1')
    dd,d0,p2 = hdr.get('CD2_2'),hdr.get('CRVAL2'),hdr.get('CRPIX2')
    if dv is None: dv = hdr.get('CDELT3')
    if dr is None: dr = hdr.get('CDELT1')
    if dd is None: dd = hdr.get('CDELT2')
    xtora = lambda x: (x-p1+1)*dr+r0    # convert pixel coordinates to RA/Dec/Velocity
    ytodec = lambda y: (y-p2+1)*dd+d0
    vconv = lambda v: (v-p3+1)*dv+v0
    if hdr.get('CUNIT3') in ['m/s','M/S']:
        conversion_factor = 1000.0
        xunits = 'km/s' # change to km/s because you're converting units
    else:
        xunits = hdr.get('CUNIT3')
        if xunits in ("hz","Hz"):
            print "Converting from Hz to GHz"
            xunits = "GHz"
            conversion_factor = 1.0e9
        else:
            conversion_factor = 1.0
    units = hdr.get('BUNIT')

    return dv,v0,p3,hdr,cube,xtora,ytodec,vconv,xunits,conversion_factor,units

def open_1d(filename,specnum=0,wcstype='',errspecnum=None,maskspecnum=None):
    """
    Grabs all the relevant pieces of a 1d spectrum for plotting
    wcstype is the suffix on the WCS type to get to velocity/frequency/whatever
    """
    f = pyfits.open(filename)
    hdr = f[0].header
    spec = f[0].data
    errspec  = None
    maskspec = None
    if hdr.get('NAXIS') == 2:
        if errspecnum is not None:
            errspec = spec[errspecnum,:]
        if maskspecnum is not None:
            maskspec = spec[maskspecnum,:]
        if isinstance(specnum,list):
            spec = spec[specnum,:].mean(axis=0)
        elif isinstance(specnum,int):
            spec = spec[specnum,:]
        else:
            raise TypeError("Specnum is of wrong type (not a list of integers or an integer).  Type: %s" %
                    str(type(specnum)))
    elif hdr.get('NAXIS') > 2:
        raise ValueError("Too many axes for open_1d (splat_1d) - use cube instead")
    if hdr.get('CD1_1'+wcstype):
        dv,v0,p3 = hdr['CD1_1'+wcstype],hdr['CRVAL1'+wcstype],hdr['CRPIX1'+wcstype]
    else:
        dv,v0,p3 = hdr['CDELT1'+wcstype],hdr['CRVAL1'+wcstype],hdr['CRPIX1'+wcstype]
    if hdr.get('OBJECT'):
        specname = hdr.get('OBJECT')
    elif hdr.get('GLON') and hdr.get('GLAT'):
        specname = "%s %s" % (hdr.get('GLON'),hdr.get('GLAT'))
    else:
        specname = filename.rstrip(".fits")
    if hdr.get('CUNIT1'+wcstype) in ['m/s','M/S']:
        conversion_factor = 1000.0
        xunits = 'km/s' # change to km/s because you're converting units
    else:
        xunits = hdr.get('CUNIT1'+wcstype)
        if xunits in ("hz","Hz"):
            print "Converting from Hz to GHz"
            xunits = "GHz"
            conversion_factor = 1.0e9
        else:
            conversion_factor = 1.0
    vconv = lambda v: ((v-p3+1)*dv+v0)/conversion_factor
    xtora=None
    ytodec=None
    units = hdr.get('BUNIT').strip()
    if hdr.get('CTYPE1'+wcstype):
        xtype = hdr.get('CTYPE1'+wcstype)
    else:
        xtype = 'VLSR'
    if hdr.get('REFFREQ'+wcstype):
        reffreq = hdr.get('REFFREQ'+wcstype)
    else:
        reffreq = None

    return dv,v0,p3,conversion_factor,hdr,spec,vconv,xtora,ytodec,specname,units,xunits,errspec,maskspec,reffreq

def splat_1d(filename=None,vmin=None,vmax=None,button=1,dobaseline=False,
        exclude=None,smooth=None,order=1,savepre=None,vcrop=True,
        vconv=None,vpars=None,hdr=None,spec=None,xtora=None,ytodec=None,
        specname=None,quiet=True,specnum=0,errspecnum=None,wcstype='',
        offset=0.0, continuum=0.0, annotatebaseline=False, plotspectrum=True,
        smoothto=None, xunits=None, units=None, conversion_factor=None,
        smoothtype='gaussian',convmode='valid',maskspecnum=None,**kwargs):
    """
    Wrapper for specplotter creation.  Works nicely with 1D spectra with well-defined
    FITS headers (i.e., CRVAL1, CRPIX1, CDELT1, and optionally CUNIT1 and CTYPE1)

    This documentation needs to be updated a lot... I implemented a lot of features
    without documenting them, which was a mistake

    Inputs:
        vmin,vmax - range over which to baseline and plot
        exclude - (internal) range to exclude from baseline fit
        vcrop - will vmin/vmax crop out data, or just set the plot limits?
    """
    if (vpars and vconv and hdr and spec is not None and xtora and ytodec 
            and units and xunits and conversion_factor):
        dv,v0,p3 = vpars
        errspec = None
        maskspec = None
        reffreq = None
        if units is None and kwargs.has_key('units'): units = kwargs.pop('units')
    else:
        dv,v0,p3,conversion_factor,hdr,spec,vconv,xtora,ytodec,specname_file,units,xunits,errspec,maskspec,reffreq = \
                open_1d(filename,specnum=specnum,wcstype=wcstype,errspecnum=errspecnum,maskspecnum=maskspecnum)
        if specname is None: specname=specname_file
        if units is None and kwargs.has_key('units'): units = kwargs.pop('units')

    if type(continuum)==type('str'):
        if hdr.get(continuum) is not None:
            continuum = hdr.get(continuum)
        else:
            raise ValueError("Continuum specified but none present.")

    varr = vconv(arange(spec.shape[0]))
    if vmin is None or vcrop==False: argvmin = 0
    else: 
      argvmin = argmin(abs(varr-vmin))
      if dv > 0:
        hdr.update('CRPIX1'+wcstype,p3-argvmin)
    if vmax is None or vcrop==False: argvmax = spec.shape[0]
    else: 
      argvmax = argmin(abs(varr-vmax))
      if dv < 0:
        hdr.update('CRPIX1'+wcstype,p3-argvmax)

    if argvmin > argvmax:
        argvmin,argvmax = argvmax,argvmin
        #if exclude is not None: exclude = exclude[::-1]
    elif argvmin == argvmax:
        raise Exception("Error: no data in velocity range %g:%g for source %s."
                % (vmin,vmax,filename))

    # these lines were meant to automatically put "exclude" into velocity
    # units; this is now done in the baseline code
    #if exclude is not None:
    #    exclude[0] = argmin(abs(varr-exclude[0]))
    #    exclude[1] = argmin(abs(varr-exclude[1]))
    #    exclude = array(exclude) - argvmin

    vconv = lambda v: ((v-p3+argvmin+1)*dv+v0) / conversion_factor
    ivconv = lambda V: p3-1-argvmin+(V*conversion_factor-v0)/dv

    specplot = spec[argvmin:argvmax]
    if errspec is not None: errspec=errspec[argvmin:argvmax]
    if maskspec is not None: maskspec=maskspec[argvmin:argvmax]

    if smoothto:
        smooth = abs(smoothto/dv)

    if smooth:
        roundsmooth = round(smooth) # can only downsample by integers
        # change fitter first
        if smoothtype == 'hanning': 
            specplot = convolve(specplot,hanning(2+roundsmooth)/hanning(2+roundsmooth).sum(),convmode)[::roundsmooth]
            kernsize = smooth
            ones_sameshape = zeros(smooth+2)
            ones_sameshape[1:-1] = 1
        elif smoothtype == 'boxcar':
            specplot = convolve(specplot,ones(roundsmooth)/float(roundsmooth),convmode)[::roundsmooth]
            kernsize = roundsmooth
            ones_sameshape = ones(roundsmooth)
        elif smoothtype == 'gaussian':
            speclen = specplot.shape[0]
            xkern  = linspace(-1*smooth,smooth,smooth*3)
            kernel = exp(-xkern**2/(2*(smooth/sqrt(8*log(2)))**2))
            kernel /= kernel.sum()
            kernsize = len(kernel)
            specplot = convolve(specplot,kernel,convmode)[::roundsmooth] 
            ones_sameshape = zeros(roundsmooth*3)
            ones_sameshape[roundsmooth:-roundsmooth] = 1
        if errspec is not None: 
            errspec = sqrt(convolve(errspec**2,ones_sameshape,convmode)[::roundsmooth]) / float(roundsmooth)
        if maskspec is not None: 
            maskspec = array(convolve(maskspec,ones_sameshape,convmode)[::roundsmooth],dtype='bool')
            if maskspec.shape != specplot.shape: import pdb; pdb.set_trace()
        # this bit of code may also make sense, but I'm shifting the center pixel instead
        # b/c it's easier (?) to deal with velocity range
        #v0 += (abs(dv)*smooth - abs(dv))/2.0 # pixel center moves by half the original pixel size
        dv *= roundsmooth
        if convmode == 'same':
          newrefpix = (p3-argvmin)/roundsmooth  
        elif convmode == 'full':
          newrefpix = (p3-0.5-argvmin+kernsize/2.0)/roundsmooth  
        elif convmode == 'valid':
          newrefpix = (p3-0.5-argvmin-kernsize/2.0)/roundsmooth  
        # this was resolved by advanced guess-and check
        # but also, sort of makes sense: FITS refers to the *center* of a pixel.  You want to 
        # shift 1/2 pixel to the right so that the first pixel goes from 0 to 1
        vconv = lambda v: ((v-newrefpix)*dv+v0)/conversion_factor
        ivconv = lambda V: newrefpix+(V*conversion_factor-v0)/dv
        hdr.update('CRPIX1'+wcstype,newrefpix+1)
        hdr.update('CDELT1'+wcstype,dv)

    sp = SpecPlotter(specplot, vconv=vconv, xtora=xtora, ytodec=ytodec,
            specname=specname, dv=dv/conversion_factor, hdr=hdr, reffreq=reffreq,
            errspec=errspec, maskspec=maskspec, xunits=xunits, **kwargs)

    if plotspectrum:
        sp.plotspec(button=button, cube=False, vmin=vmin, vmax=vmax,
                units=units, offset=offset, continuum=continuum,
                **kwargs)

    if dobaseline: 
        sp.baseline(exclude=exclude,order=order,quiet=quiet,annotate=annotatebaseline)
    if plotspectrum: sp.refresh()
    
    if hdr.get('GLON') and hdr.get('GLAT'):
        sp.glon = hdr.get('GLON')
        sp.glat = hdr.get('GLAT')

    if savepre is not None:
        glon,glat = sp.glon,sp.glat
        if glat < 0: pm="" 
        else: pm = "+"
        savename = savepre + "G%07.3f%0s%07.3f_" % (glon,pm,glat) + hdr.get('MOLECULE').replace(' ','') + hdr.get('TRANSITI').replace(' ','')
        savefig(savename+'.png')

    return sp

def splat_tspec(filename,specnum=0,**kwargs):
    """
    Same as splat_1d for tspec data
    """

    tdata = pyfits.getdata(filename)
    theader = pyfits.getheader(filename)
    if len(tdata.shape) == 3:
        tdata = tdata[specnum,:,:]
    wavelength = tdata[0,:]
    spectrum   = tdata[1,:]
    error      = tdata[2,:]

    vconv = lambda x: wavelength[x]
    ivconv = lambda x: argmin(abs(wavelength-x))
    
    specname='TSPEC'
    dv = median(wavelength[1:] - wavelength[:-1])
    
    sp = SpecPlotter(spectrum,vconv=vconv,specname=specname,dv=dv,hdr=theader)

    sp.plotspec(cube=False,units=theader.get('YUNITS'),xunits=theader.get('XUNITS'),**kwargs)

    return sp



import pyfits

def intersection(header1, header2, if_conflict=None):
    """
    Return a pyfits Header containing the intersection of two pyfits Headers

    *if_conflict* [ '1'/1/'Header1' | '2'/2/'Header2' | None ]
        Defines behavior if a keyword conflict is found.  Default is to remove the key

    """

    newheader = pyfits.Header()

    for key,value in header1.items():
        if key in header2:
            try:
                if value == header2[key]:
                    newheader[key] = value
                elif if_conflict in ('1',1,'Header1'):
                    newheader[key] = value
                elif if_conflict in ('2',2,'Header2'):
                    newheader[key] = Header2[key]
            except KeyError:
                """ Assume pyfits doesn't want you to have that keyword
                (because it shouldn't be possible to get here otherwise) """
                pass
        else:
            try:
                newheader[key] = value
            except KeyError:
                """ Assume pyfits doesn't want you to have that keyword
                (because it shouldn't be possible to get here otherwise) """
                pass

    return newheader



import os
from . import Writer

h5check = True
try: import h5py
except ImportError: h5check = False

class write_hdf5(Writer):
        
    def write_data(self, filename = None, newsuffix = 'out', 
        extras = None, clobber = True):
        """
        Write information to hdf5 file.
        
        extras = [(dataset name, data), (dataset name, data)] 
        
        To do: leave option of writing to groups (for model components especially?)
        """
        
        if not h5check: print "Cannot write to hdf5 - h5py import failed."
        
        else: 
            if filename is None: fn = "{0}_{1}.hdf5".format(self.Spectrum.fileprefix, newsuffix)
            else: fn = filename
            
            if clobber: f = h5py.File(fn, 'w')
            else:
                i = 1
                while os.path.exists(fn):
                    fn = "{0}_fit({1}).hdf5".format(self.Spectrum.fileprefix, i)
                    i += 1
                            
            # By default, write out x-array, spectrum, errors
            xarr = f.create_dataset('xarr', data = self.Spectrum.xarr)
            data = f.create_dataset('data', data = self.Spectrum.data)
            error = f.create_dataset('error', data = self.Spectrum.error)
            
            # Add metadata to each dataset?
            xarr.attrs.create('units', self.Spectrum.xarr.units)
            xarr.attrs.create('type', self.Spectrum.xarr.xtype)
            data.attrs.create('units', self.Spectrum.units)
            data.attrs.create('type', self.Spectrum.ytype)
            
            if extras is not None:
                for extra in extras:
                    f.create_dataset(extra[0], data = extra[1]) # add try-except
                
            f.close()
            
